package edu.ntnu.item.jt.VMware;

import com.sun.jna.Callback;

import edu.ntnu.item.jt.VMware.*;

/**
 * A simple generic callback procedure which accepts a callback, 
 * responds on queries of whether the callback has been received and 
 * releases the job from the Vix API upon completion.
 * 
 * Callbacks are used when executing operations in the Vix API in a non-blocking manner.
 * The Vix API (via JNA) will call the provided callback when the given operation is finished. 
 * 
 * @author Jran Vagnby Lillesand
 *
 * @see com.sun.jna.Callback
 */
public class CallbackProc implements Callback {
	
	private GuestOS guestOS;
	private boolean operationFinished;
	private int jobHandle = -1;
	
	/**
	 * Creates a new CallbackProc instance.
	 * 
	 * Takes a reference to the guest OS on which the job is being executed.
	 * This is later used to free up the resources associated with the job. 
	 * 
	 * 
	 * @param guestOS
	 */
	public CallbackProc(GuestOS guestOS){
		this.guestOS = guestOS;
		operationFinished = false;
	}
	
	/**
	 * Sets the job handle associated with this job.
	 * This should be done at some time before the object is finalized, at which point the job handle is used to free up the resources allocated to this job by the Vix API. 
	 * 
	 * @param jobHandle handle of the job that is being executed.
	 */
	public void setJobHandle(int jobHandle){
		this.jobHandle = jobHandle;
	}
	
	/**
	 * Called by JNA/Vix when the operation is completed.
	 */
	public void callback(){
		operationFinished = true;
	}
	
	/**
	 * Check whether the callback has been received.
	 * 
	 * @return true if the operation has finished, false if not.
	 */
	public boolean isFinished(){
		return operationFinished;
	}
	
	/**
	 * Unregister the job with the Vix API to free up any reserved resources.
	 */
	protected void finalize(){
		// Unregister job.
		if(jobHandle > 0){
			guestOS.releaseJobHandle(jobHandle);
		}
	}
}
