package edu.ntnu.item.jt.VMware;

import edu.ntnu.item.jt.VMware.errors.*;

import java.util.logging.*;
import java.util.Date;
import java.io.IOException;

import com.sun.jna.*;

/**
 * Represents a Guest OS/Virtual Machine instance.
 * 
 * Instances of GuestOS can only be gotten through a relevant VMwareServer, as a virtual machine always
 * must be directly associated with a VMware server. The GuestOS is aware of its VMware server, but not vice versa.
 * 
 * There is state associated with the virtual machine (GuestOS), but this is not reflected in the GuestOS class, as
 * the underlying VIX API will return appropriate errors (which will be represented by exceptions here) if the state is wrong for a given call.
 * 
 * Most functionality offered in this class utilizes in turn the VixInterface class and the C VIX API offered by VMWare.
 *  
 * @author Jran Vagnby Lillesand
 *
 */
public class GuestOS {

	/*
	 * The following defines the relevant constants from the vix.dll API.
	 */
	private static final int VIX_VMPOWEROP_NORMAL 					= 0; 
	private static final int VIX_INVALID_HANDLE 					= 0;
	public static final int VIX_PROPERTY_NONE 						= 0;
	public static final long VIX_OK									= 0;
	public static final int VIX_SNAPSHOT_INCLUDE_MEMORY 			= 0x0002;
	public static final int VIX_PROPERTY_JOB_RESULT_HANDLE 			= 3010;
	public static final int VIX_VM_POWEROP_NORMAL 					= 0;
	public static final int VIX_E_FILE_ALREADY_EXISTS				= 12;
	public static final int VIX_RUNPROGRAM_RETURN_IMMEDIATELY   	= 0x0001;
	public static final int VIX_RUNPROGRAM_ACTIVATE_WINDOW      	= 0x0002;
	public static final String VIX_CONSOLE_USER_NAME          		= "__VMware_Vix_Guest_Console_User__";

	
	private int[] vmHandle;
	private int[] snapshotHandle;
	private VMwareServer vmWareServer;
	private VixInterface vixLib;
	private String vmxPath;
	private String username;
	private String password;
	
	private static Logger log = Logger.getLogger("edu.ntnu.item.jt.vmware");
	
	/**
	 * Constructor for the GuestOS class.
	 * 
	 * The constructor is protected, as guest OS's may only be returned via the appropriate VMwareServer object.
	 * 
	 * @param vmHandle the Vix handle of this virtual machine.
	 * @param vmxPath the VMX config path of this virtual machine. The VMX path is valid on the VMware Server running this guest OS.
	 * @param vmWareServer reference to the VMware server running this guest OS instance.
	 * @param username username for logging in to the OS running at the virtual machine.
	 * @param password password for logging in to the OS running at the virtual machine. If no password is required, supply null.
	 */
	protected GuestOS(int[] vmHandle, String vmxPath, String username, String password, VMwareServer vmWareServer){
		vixLib = VixInterface.INSTANCE;
		this.vmHandle = vmHandle;
		this.vmWareServer = vmWareServer;
		this.vmxPath = vmxPath;
		this.username = username;
		this.password = password;
		
	}
	
	/**
	 * Powers on the virtual machine. If the virtual machine is already powered on,
	 * the function proceeds as normal (returns).
	 * 
	 * The function can be instructed to wait for VMWare tools to respond in the guest OS.
	 * This can be done using a timeout which will be triggered if Tools does not respond within
	 * the given timeout.
	 * 
	 * @param waitForTools specifies whether or not to wait for tools to run in the guest os.
	 * @param timeout specifies the timeout in seconds when waiting for tools. If waitForTools is false, this parameter is ignored.
	 * @throws PowerUpFailureException the only exception thrown. Contains a textual description of the error that occured.
	 */
	public void powerOn(boolean waitForTools, int timeout) throws PowerUpFailureException{
		log(LOG_INFO, "Attempting to power on");
		int jobHandle = vixLib.VixVM_PowerOn(vmHandle[0], VIX_VMPOWEROP_NORMAL, VIX_INVALID_HANDLE, null, null);
		
		// Wait for job to finish.
		long result = vixLib.VixJob_Wait(jobHandle, VIX_PROPERTY_NONE, null, VIX_PROPERTY_NONE);
		
		if(result != VIX_OK){
			log(LOG_WARNING, "Power on failed. Error code "+result+". Error description: "+vixLib.Vix_GetErrorText(result, null));
            vixLib.Vix_ReleaseHandle(jobHandle);
            throw new PowerUpFailureException(vixLib.Vix_GetErrorText(result, null), result);
		}else{
			if(VMwareServer.debug) System.out.println("Successfully powered up VM "+vmHandle[0]);
			vixLib.Vix_ReleaseHandle(jobHandle);
			log(LOG_INFO, "Successfully powered on");
			
			if(waitForTools){
				log(LOG_INFO, "Waiting for WMWare tools");
				if(VMwareServer.debug) System.out.println("Waiting "+timeout+ " seconds for VMWare tools in guest.");
				jobHandle = vixLib.VixVM_WaitForToolsInGuest(vmHandle[0], timeout, null, null);
				
				result = vixLib.VixJob_Wait(jobHandle, VIX_PROPERTY_NONE, null, VIX_PROPERTY_NONE);
				
				if(result != VIX_OK){
					log(LOG_WARNING, "Wait for tools failed. Error code "+result+". Error description: "+vixLib.Vix_GetErrorText(result, null));
					if(VMwareServer.debug) System.err.println("Error while waiting for tools: "+vixLib.Vix_GetErrorText(result, null));
					throw new PowerUpFailureException("Tools did not respond before timeout. Vix error: "+vixLib.Vix_GetErrorText(result, null), result);
					
				}else{
					log(LOG_INFO, "Tools ready.");
					if(VMwareServer.debug) System.out.println("Tools ready.");
				}
				vixLib.Vix_ReleaseHandle(jobHandle);
			}
		}
	}
	
	/**
	 * Powers off the virtual machine.
	 * 
	 * @throws PowerOffFailureException if an error occured while powering off the virtual machine.  
	 */
	public void powerOff() throws PowerOffFailureException{
		log(LOG_INFO, "Attempting to power off");

		int jobHandle = vixLib.VixVM_PowerOff(vmHandle[0], 0, null, null);
		
		long result = vixLib.VixJob_Wait(jobHandle, VIX_PROPERTY_NONE, null, VIX_PROPERTY_NONE);
		
		if(result != VIX_OK){
			log(LOG_WARNING, "Powering off failed. Error code "+result+". Error description: "+vixLib.Vix_GetErrorText(result, null));

            vixLib.Vix_ReleaseHandle(jobHandle);
            throw new PowerOffFailureException(vixLib.Vix_GetErrorText(result, null), result);
		}else{
			if(VMwareServer.debug) System.out.println("Successfully powered off VM "+vmHandle[0]);
			log(LOG_INFO, "Successfully powered off.");
			vixLib.Vix_ReleaseHandle(jobHandle);
		}
	}

	/**
	 * Login to the guest OS. Note that it is required that the user account has an associated password (only verified for Win XP).
	 * 
	 * If the virtual machine is not powered on, an exception will be thrown.
	 * 
	 * @throws LoginFailureException indicates an error during logging in. Contains a textual description of the error.
	 */
	public void login() throws LoginFailureException{
		log(LOG_INFO, "Attempting to login with "+this.username+"/"+this.password+".");

		int jobHandle = vixLib.VixVM_LoginInGuest(vmHandle[0], username, password, 0, null, null);
		
		long result = vixLib.VixJob_Wait(jobHandle, VIX_PROPERTY_NONE, null, VIX_PROPERTY_NONE);
		
		if(result != VIX_OK){
			log(LOG_WARNING, "Login failed. Error code "+result+". Error description: "+vixLib.Vix_GetErrorText(result, null));

            vixLib.Vix_ReleaseHandle(jobHandle);
            throw new LoginFailureException(vixLib.Vix_GetErrorText(result, null), result);
		}else{
			if(VMwareServer.debug) System.out.println("Logged in on VM "+vmHandle[0]+" as "+username+".");
			log(LOG_INFO, "Successfully logged in.");

			vixLib.Vix_ReleaseHandle(jobHandle);
		}
	}

	/**
	 * Copy a file to the guest OS. Note that the file is copied from the client running the code (this code),
	 * not the VMWare host.
	 * 
	 * The call blocks until file transfer is completed.
	 * 
	 * @param localPathName full path of the local file. 
	 * @param guestPathName full path of the destination file on the guest OS.
	 * @throws FileCopyException thrown if an error occurs during copying of the file. Includes a textual representation of the error.
	 */
	public void copyFileToGuest(String localPathName, String guestPathName) throws FileCopyException{
		int jobHandle = vixLib.VixVM_CopyFileFromHostToGuest(vmHandle[0], localPathName, guestPathName, 0, VIX_INVALID_HANDLE, null, null);
		log(LOG_INFO, "Attempting to copy file from local path "+localPathName+" to guest OS path "+guestPathName+".");

		long result = vixLib.VixJob_Wait(jobHandle, VIX_PROPERTY_NONE, null, VIX_PROPERTY_NONE);
		
		if(result != VIX_OK){
            vixLib.Vix_ReleaseHandle(jobHandle);
			log(LOG_WARNING, "File copy failed. Error code "+result+". Error description: "+vixLib.Vix_GetErrorText(result, null));

            throw new FileCopyException(vixLib.Vix_GetErrorText(result, null), result);
		}else{
			if(VMwareServer.debug) System.out.println("Copied file from path "+localPathName+" on host OS to path "+guestPathName+" on VM "+ vmHandle[0]+".");
			log(LOG_INFO, "Copied file from local path "+localPathName+" to guest OS path "+guestPathName+".");

			vixLib.Vix_ReleaseHandle(jobHandle);
		}
	}
	
	/**
	 * Copy a file from the local machine (client) to the guest OS.
	 * 
	 * The call blocks until file transfer is completed.
	 * 
	 * @param guestPathName file name of the file in the guest OS. 
	 * @param hostPathName file name of destination file in the local machine.
	 * @throws FileCopyException thrown if an error occurs during copying of the file. Includes a textual representation of the error.
	 */
	public void copyFileFromGuest(String guestPathName, String hostPathName) throws FileCopyException{
		log(LOG_INFO, "Attempting to copy file from guest OS path "+guestPathName+" to local path "+hostPathName+".");

		int jobHandle = vixLib.VixVM_CopyFileFromGuestToHost(vmHandle[0], guestPathName, hostPathName, 0, VIX_INVALID_HANDLE, null, null);
		
		long result = vixLib.VixJob_Wait(jobHandle, VIX_PROPERTY_NONE, null, VIX_PROPERTY_NONE);
		
		if(result != VIX_OK){
            vixLib.Vix_ReleaseHandle(jobHandle);
        	log(LOG_WARNING, "File copy failed. Error code "+result+". Error description: "+vixLib.Vix_GetErrorText(result, null));
            
            throw new FileCopyException(vixLib.Vix_GetErrorText(result, null), result);
		}else{
			log(LOG_INFO, "Copied from guest OS path "+guestPathName+" to local path "+hostPathName+".");

			if(VMwareServer.debug) System.out.println("Copied file from path "+guestPathName+" on VM "+vmHandle[0]+" to path "+hostPathName+" on host OS.");
			vixLib.Vix_ReleaseHandle(jobHandle);
		}
	}
	
	/**
	 * Run a executable file in the guest OS.
	 * 
	 * In order to read from the program's stdout stream, pipe the output to a file using arguments 
	 * and copy this file using the copyFileFromGuest function:
	 * 
	 * guestOS.runProgramInGuest("/bin/ls", "/bin > /tmp/ls.out");
	 * guestOs.copyFileFromGuest("/tmp/ls.out", "c:\\ls.out");
	 * 
	 * @param programPath full path of the program to be executed.
	 * @param commandLineArgs command line arguments to be supplied to the executable.
	 * @throws FileExecutionException thrown if an error occured when executing the file.
	 */
	public void runProgramInGuest(String programPath, String commandLineArgs) throws FileExecutionException{
		log(LOG_INFO, "Executing "+programPath+" with parameters "+commandLineArgs);
		
		int jobHandle = vixLib.VixVM_RunProgramInGuest(vmHandle[0], programPath, commandLineArgs, 0, VIX_INVALID_HANDLE, null, null);
		
		long result = vixLib.VixJob_Wait(jobHandle, VIX_PROPERTY_NONE, null, VIX_PROPERTY_NONE);
		
		if(result != VIX_OK){
            vixLib.Vix_ReleaseHandle(jobHandle);
            
            log(LOG_WARNING, "File execution failed. Error code "+result+". Error description: "+vixLib.Vix_GetErrorText(result, null));
            throw new FileExecutionException(vixLib.Vix_GetErrorText(result, null), result);
		}else{
			log(LOG_INFO, "Successfully executed "+programPath+".");
			
			if(VMwareServer.debug) System.out.println("Successfully executed "+programPath+ " on VM "+vmHandle[0]+".");
			vixLib.Vix_ReleaseHandle(jobHandle);
		}
	}

	/**
	 * Log in to the Guest OS with a GUI console user. This causes running programs to be visible when the logging in
	 * to the VM using the console. The program is run within the context of the user stated in the previous log in operation.
	 * 
	 * @throws LoginFailureException If the log in operation fails
	 */
	public void logInGuiConsoleUser() throws LoginFailureException {
		
		log(LOG_INFO,"Attempting login with GUI console user");
		
		//Log in with "console" user
		int jobHandle = vixLib.VixVM_LoginInGuest(vmHandle[0], VIX_CONSOLE_USER_NAME, null, 0, null, null);
		
		long result = vixLib.VixJob_Wait(jobHandle, VIX_PROPERTY_NONE, null, VIX_PROPERTY_NONE);
		
		if(result != VIX_OK){
            vixLib.Vix_ReleaseHandle(jobHandle);
            
            log(LOG_WARNING, "Login with GUI console user failed. Error description: "+vixLib.Vix_GetErrorText(result, null));
            throw new LoginFailureException(vixLib.Vix_GetErrorText(result, null), result);
		}else{
			log(LOG_INFO, "Successfully logged in with GUI console user.");
			
			if(VMwareServer.debug) System.out.println("Successfully logged in with GUI console user.");
			vixLib.Vix_ReleaseHandle(jobHandle);
		}				
		
	}
	
	/**
	 * Run a program in the guest OS without blocking call or any callback procedure.
	 * 
	 * NB! When using this method, there is no way to evaluate the result status code of the call!
	 * 
	 * @param programPath Path to the program to be executed
	 * @param commandLineArgs any command line arguments
	 */
	public void runProgramInGuestNonBlocking(String programPath, String commandLineArgs) throws FileExecutionException{
		log(LOG_INFO, "Executing "+programPath+" with parameters "+commandLineArgs+" as a non-blocking call");
		
		int jobHandle2 = vixLib.VixVM_RunProgramInGuest(vmHandle[0], programPath, commandLineArgs, VIX_RUNPROGRAM_RETURN_IMMEDIATELY | VIX_RUNPROGRAM_ACTIVATE_WINDOW , VIX_INVALID_HANDLE, null, null);

		long result2 = vixLib.VixJob_Wait(jobHandle2, VIX_PROPERTY_NONE, null, VIX_PROPERTY_NONE);
		
		if(result2 != VIX_OK){
            vixLib.Vix_ReleaseHandle(jobHandle2);
            
            log(LOG_WARNING, "File execution failed. Error code "+result2+". Error description: "+vixLib.Vix_GetErrorText(result2, null));
            throw new FileExecutionException(vixLib.Vix_GetErrorText(result2, null), result2);
		}else{
			log(LOG_INFO, "Successfully executed "+programPath+".");
			
			if(VMwareServer.debug) System.out.println("Successfully executed "+programPath+ " on VM "+vmHandle[0]+".");
			vixLib.Vix_ReleaseHandle(jobHandle2);
		}
		
	}
	
	/**
	 * Run a program in the guest OS using a callback procedure. 
	 * 
	 * The callback procedure will be called upon finishing the operation.
	 * 
	 * Note that to ensure that the program has finished executing the callback procedure class should be polled.
	 * 
	 * @param programPath path to the program to be executed. 
	 * @param commandLineArgs any command line arguments to be passed to the program.
	 * @param callbackProc class containing the callback procedure.
	 * @return job handle of the job (needed to free up the resources allocated by vix when executing the program). 
	 */
	public int runProgramInGuest(String programPath, String commandLineArgs, Callback callbackProc) {
		log(LOG_INFO, "Executing "+programPath+" with parameters "+commandLineArgs+ " using callback.");
		
		int jobHandle = vixLib.VixVM_RunProgramInGuest(vmHandle[0], programPath, commandLineArgs, 0, VIX_INVALID_HANDLE, callbackProc, null);
		
		return jobHandle;
	}
	
	/**
	 * Take a snapshot of the current state in the virtual machine (guest OS). May also include the current
	 * state of the memory in the guest. Note that this implementation is written for VMWare server, which only
	 * supports one snapshot at the time. This means that if a new snapshot is taken, any previously stored one 
	 * will be overwritten.
	 * 
	 * @param name a name describing the snapshot. Only used for description, so need not be unique.
	 * @param description a textual description of the snapshot.
	 * @param includeMemory indicates whether the memory state is to be part of the snapshot.
	 * @throws CreateSnapshotException an error occured while trying to create the snapshot. See the textual description and/or the exception error code for more information.
	 */
	public void takeSnapshot(String name, String description, boolean includeMemory) throws CreateSnapshotException{
		log(LOG_INFO, "Attempting to capture snapshot with name "+name+" and description "+description+". Include memory i snapshot: "+includeMemory);

		int jobHandle = 0;
		if(includeMemory){
			jobHandle = vixLib.VixVM_CreateSnapshot(vmHandle[0], name, description, VIX_SNAPSHOT_INCLUDE_MEMORY, VIX_INVALID_HANDLE, null, null);
		}else{
			jobHandle = vixLib.VixVM_CreateSnapshot(vmHandle[0], name, description, 0, VIX_INVALID_HANDLE, null, null);
		}
		
		snapshotHandle = new int[1];
		long result = vixLib.VixJob_Wait(jobHandle, VIX_PROPERTY_JOB_RESULT_HANDLE, snapshotHandle, VIX_PROPERTY_NONE);
		
		if(result != VIX_OK){
			snapshotHandle = null;
			log(LOG_WARNING, "Snapshot capture failed. Error code "+result+". Error description: "+vixLib.Vix_GetErrorText(result, null));
            vixLib.Vix_ReleaseHandle(jobHandle);
            throw new CreateSnapshotException(vixLib.Vix_GetErrorText(result, null), result);
		}else{
			log(LOG_INFO, "Successfully created snapshot " +name);

			if(VMwareServer.debug) System.out.println("Successfully created snapshot with handle "+snapshotHandle[0]+".");
			vixLib.Vix_ReleaseHandle(jobHandle);
		}
	}
	
	/**
	 * Revert to the previous snapshot. First attempts to load the snapshot, and the reverts to it.
	 * 
	 * @throws NoSnapshotLoadedException indicates that no snapshot has been loaded prior to calling the function.
	 * @throws RevertToSnapshotFailedException indicates that an error occured while attempting to revert to the loaded snapshot. For information about the error, see the exception's textual description or error code.
	 * @throws NoSnapshotAvailableException no snapshot could be loaded, indicating that no snapshot is available.
	 */
	public void revertToSnapshot(boolean waitForCompletion) throws NoSnapshotLoadedException, RevertToSnapshotFailedException, NoSnapshotAvailableException{
		if(snapshotHandle == null){
			log(LOG_INFO, "Attempted to revert to snapshot while none was loaded.");
			loadSnapshot();
			//throw new NoSnapshotLoadedException("No snapshot handle loaded. Snapshot must be created or loaded prior to calling.", 0);
		}
		
		log(LOG_INFO, "reverting to snapshot "+snapshotHandle[0]);

		//TODO: This does not seem to work if too short time lapses between the previous create snapshot and the next load. wrong handle is returned.
		//loadSnapshot();		
		
		int jobHandle = vixLib.VixVM_RevertToSnapshot(vmHandle[0], snapshotHandle[0], VIX_VM_POWEROP_NORMAL, VIX_INVALID_HANDLE, null, null);
		
		if(waitForCompletion){
			long result = vixLib.VixJob_Wait(jobHandle, VIX_PROPERTY_NONE, null, VIX_PROPERTY_NONE);
			
			
			if(result != VIX_OK){
				log(LOG_WARNING, "Reverting to snapshot failed. Error code "+result+". Error description: "+vixLib.Vix_GetErrorText(result, null));
				
	            vixLib.Vix_ReleaseHandle(jobHandle);
	            throw new RevertToSnapshotFailedException(vixLib.Vix_GetErrorText(result, null), result);
			}else{
				log(LOG_INFO, "Successfully reverted to snapshot "+snapshotHandle[0]);

				if(VMwareServer.debug) System.out.println("Successfully reverted to snapshot with handle "+snapshotHandle[0]+" in VM " +vmHandle[0]+ ".");
				vixLib.Vix_ReleaseHandle(jobHandle);
			}
		}else{
			log(LOG_INFO, "Reverted to snapshot without waiting for operation to finish. Result status unknown.");
		}
	}
			
	
	/**
	 * Loads a previously stored snapshot. Does not perform any changes to the VM, but loads
	 * the handle of the snapshot 
	 * 
	 * @throws NoSnapshotLoadedException
	 * @throws NoSnapshotAvailableException
	 */
	public void loadSnapshot() throws NoSnapshotLoadedException, NoSnapshotAvailableException{
		log(LOG_INFO, "Attempting to load snapshot.");
		
		// Create container for result of VixVM_GetNumRootSnapshots
		int[] numRootSnapshots = new int[1];
		
		// Check the number of available snapshots (either 0 or 1 for VMWare server).
		long result = vixLib.VixVM_GetNumRootSnapshots(vmHandle[0], numRootSnapshots);
		
		if(result != VIX_OK)
            throw new NoSnapshotLoadedException(vixLib.Vix_GetErrorText(result, null), result);
		
		if(VMwareServer.debug) System.out.println("Number of snapshots in VM "+vmHandle[0]+":"+numRootSnapshots[0]);
		
		if(numRootSnapshots[0] == 0)
			throw new NoSnapshotAvailableException();
		
		//VMWare server only supports one snapshot, which will be at index 0
		snapshotHandle = new int[1];
		result = vixLib.VixVM_GetRootSnapshot(vmHandle[0], 0, snapshotHandle);
		
		if(result != VIX_OK){
			log(LOG_WARNING, "Loading of snapshot failed. Error code "+result+". Error description: "+vixLib.Vix_GetErrorText(result, null));
            snapshotHandle = null;
            throw new NoSnapshotLoadedException(vixLib.Vix_GetErrorText(result, null), result);
		}
		
		log(LOG_INFO, "Successfully loaded snapshot handle "+snapshotHandle[0]);
		
		if(VMwareServer.debug) System.out.println("Loaded snapshot with handle " +snapshotHandle[0]+ "in VM " +vmHandle[0]+ ".");
		//no need to release handles here, as none of these jobs are resource consuming.
	}
	
	/**
	 * Releases the given job handle in the vix API.
	 * 
	 * @param jobHandle id of the job handle to be released.
	 */
	public void releaseJobHandle(int jobHandle){
		vixLib.Vix_ReleaseHandle(jobHandle);
	}
	
	/** Not currently supported in VIX 1.0 and 1.1.
	 * 
	public void createDirectory(String directory) throws DirectoryCreationException{
		log(LOG_INFO, "Creating directory: "+directory);
		
		int jobHandle = vixLib.VixVM_CreateDirectoryInGuest(vmHandle[0], directory, VIX_PROPERTY_NONE, null, null);
		
		long result = vixLib.VixJob_Wait(jobHandle, VIX_PROPERTY_NONE, null, VIX_PROPERTY_NONE);
		
		if(result == VIX_E_FILE_ALREADY_EXISTS){
			// Do nothing. Directory exist - no problem for this usage.
			log(LOG_INFO, "Directory already exists.");
			vixLib.Vix_ReleaseHandle(jobHandle);
			return;
		}else if(result != VIX_OK){
			// Creating directory failed.
			vixLib.Vix_ReleaseHandle(jobHandle);
			log(LOG_WARNING, "Directory creation failed. Error code: "+result+". Error description: "+vixLib.Vix_GetErrorText(result, null));
			throw new DirectoryCreationException(vixLib.Vix_GetErrorText(result, null), result);
		}
	}
	*/
	
	public int getVMHandle() {
		return vmHandle[0];
	}
	
	/**
	 * Getter for the guest OS's VMX path.
	 *  
	 * @return vmx path of the guest OS.
	 */
	public String getvmxPath(){
		return vmxPath;
	}
	
	/**
	 * Getter for the hostname of the VMware server running this guest OS.
	 * 
	 * @return hostname of the VMware server hosting the guest OS.
	 */
	public String getHost(){
		return vmWareServer.getHost();
	}
	
	/**
	 * Getter for the username associated with this virtual machine. 
	 * 
	 * @return username associated with the virtual machine OS.
	 */
	public String getUsername(){
		return this.username;
	}
	
	/**
	 * Getter for the password associated with this virtual machine. 
	 * 
	 * @return password associated with the virtual machine OS.
	 */
	public String getPassword(){
		return this.password;
	}

	/**
	 * Releases the vix handle of this guest OS, allowing vix to free up any resources locked by this VM. 
	 */
	protected void finalize(){
		//attempt to free up resources allocated to VM.
		vixLib.Vix_ReleaseHandle(vmHandle[0]);
	}

	/**
	 * Private logger to assist in parsing log messages .
	 * 
	 * @param severity severity level of the occurence.
	 * @param message log message.
	 */
	private void log(int severity, String message){
		int hostHandle = vmWareServer.getHostHandle();
		
		switch(severity){
		case LOG_INFO:
			log.info("VMWare host: "+hostHandle+", VM handle:" +vmHandle[0]+": " +message);
			break;
		case LOG_WARNING:
			log.warning("VMWare host: "+hostHandle+", VM handle:" +vmHandle[0]+": " +message);
			break;
		case LOG_CONFIG:
			log.config("VMWare host: "+hostHandle+", VM handle:" +vmHandle[0]+": " +message);
			break;
		case LOG_SEVERE:
			log.severe("VMWare host: "+hostHandle+", VM handle:" +vmHandle[0]+": " +message);
			break;
		case LOG_ENTERING:
			//TODO: Implement this?
			break;
		case LOG_EXITING:
			//TODO: Implement this?
			break;
		}
	}
	
	/*
	 * Log levels
	 */
	private static final int LOG_INFO = 0;
	private static final int LOG_WARNING = 1;
	private static final int LOG_CONFIG = 2;
	private static final int LOG_SEVERE = 3;
	private static final int LOG_ENTERING = 4;
	private static final int LOG_EXITING = 5;
	
	

	
}