package edu.ntnu.item.jt.VMware;

import java.util.logging.Logger;
import edu.ntnu.item.jt.VMware.errors.*;

/**
 * Representents a VMware server.
 * 
 * @author Jran Vagnby Lillesand
 *
 */
public class VMwareServer {
	
	/*
	 * VIX constant definitions.
	 */
    public static final int VIX_API_VERSION                         = 3;
    public static final int VIX_SERVICEPROVIDER_DEFAULT             = 1;
    public static final int VIX_SERVICEPROVIDER_VMWARE_SERVER       = 2;
    public static final int VIX_SERVICEPROVIDER_VMWARE_WORKSTATION  = 3;
    public static final int VIX_INVALID_HANDLE                      = 0;
    public static final int VIX_PROPERTY_NONE                       = 0;
    public static final int VIX_PROPERTY_JOB_RESULT_HANDLE          = 3010;
    public static final int VIX_VMPOWEROP_NORMAL                    = 0;
    public static final long VIX_OK                                 = 0;
    
    private String hostname;
    private int portNumber;
    
    private int[] hostHandle = new int[1];
    
    public static boolean debug = false;
    
    VixInterface vixLib;
    
    private static Logger log = edu.ntnu.item.jt.system.PowerScan.getLogger(edu.ntnu.item.jt.configloader.PropertiesReader.getInstance().getString("log.vmware", "logs"+System.getProperty("file.separator")+"vmware.log"), "edu.ntnu.item.jt.vmware");
    
    
    /**
     * Create a new VMwareServer instance. Initializes the Vix interface.
     * 
     * @param hostname host running VMware server.
     */
    public VMwareServer(String hostname, int portNumber){
    	// Init JNA
    	vixLib = VixInterface.INSTANCE;
    	
    	this.portNumber = portNumber;
    	this.hostname = hostname;
    }
    
    /**
     * Connect to a VMWare Server.
     * 
     * To connect to server running on local machine as current user, supply null for serverUsername and serverPassword.
     * 
     * @param serverUsername user name of the user running VMWare.
     * @param serverPassword password of the user running VMWare.
     * @throws ConnectionFailedException thrown to indicate a connection failure.
     */
	public void connect(String serverUsername, String serverPassword) throws ConnectionFailedException{
		log(LOG_INFO, "Attempting to connect to virtual machine at "+hostname+" with username "+serverUsername+".");
    	int jobHandle = vixLib.VixHost_Connect(1, VIX_SERVICEPROVIDER_VMWARE_SERVER, hostname , portNumber, serverUsername, serverPassword, 0, VIX_INVALID_HANDLE, null, null);
    	
    	// Wait for connection operation to finish
 		long result = vixLib.VixJob_Wait(jobHandle, VIX_PROPERTY_JOB_RESULT_HANDLE, hostHandle, VIX_PROPERTY_NONE);
        
 		// Evaluate the return value of VixJob_Wait()
        if (result != VIX_OK) {
        	log(LOG_WARNING, "Connection to VMWare Server failed. Error code: "+result+". Textual description: "+vixLib.Vix_GetErrorText(result, null));
        	
            vixLib.Vix_ReleaseHandle(jobHandle);
            throw new ConnectionFailedException(vixLib.Vix_GetErrorText(result, null), result);
        }else{
        	log(LOG_INFO, "Successfully connected to VMWare server. Handle: "+hostHandle[0]);
        	if(debug) System.out.println("Successfully connected to VMWare server.");
        	vixLib.Vix_ReleaseHandle(jobHandle);
        }
	}
	
	/**
	 * Load a VM in the VMWare server.
	 * 
	 * Requires that a connection has already been established to a VMWare server.
	 * 
	 * @param vmxFilePathName Full path name to the .vmx VMWare config file on the VMWare server host.
	 * @param username user name for the virtual machine.
	 * @param password password for the virtual machine. If none is required, supply null.
	 * @return instance representing the loaded guest OS.
	 * @throws OpenVMException thrown if opening the VM failed.
	 * @see GuestOS
	 */
	public GuestOS openVM(String vmxFilePathName, String username, String password) throws OpenVMException{
		log(LOG_INFO, "Attempting to open virtual machine with config file at path "+vmxFilePathName);
		
		int jobHandle = vixLib.VixVM_Open(hostHandle[0], vmxFilePathName, null, null);
		int[] vmHandle = new int[1];
		
		// Wait for operation to finish
		long result = vixLib.VixJob_Wait(jobHandle, VIX_PROPERTY_JOB_RESULT_HANDLE, vmHandle, VIX_PROPERTY_NONE);
		
		if(result != VIX_OK){
			log(LOG_WARNING, "Opening virtual machine failed. Error code: "+result+". Textual description: "+vixLib.Vix_GetErrorText(result, null));
		
            vixLib.Vix_ReleaseHandle(jobHandle);
            throw new OpenVMException (vixLib.Vix_GetErrorText(result, null), result);
		}else{
			log(LOG_INFO, "Successfully opened virtual machine. VM handle: "+vmHandle[0]);
			if(debug) System.out.println("Successfully connected to VM "+vmxFilePathName);
			vixLib.Vix_ReleaseHandle(jobHandle);
			return new GuestOS(vmHandle, vmxFilePathName, username, password, this);
		}
		
	}
	
	/**
	 * Disconnect from the guest OS and free the resources.
	 */
	public void disconnect(){
		log(LOG_INFO, "Disconnecting from VMWare server.");
		vixLib.VixHost_Disconnect(hostHandle[0]);
		hostHandle = null;
	}
	
	protected void finalize(){
		disconnect();
	}
	
	/**
	 * Private logger to assist in parsing log messages .
	 * 
	 * @param severity severity level of the occurence.
	 * @param message log message.
	 */
	private void log(int severity, String message){
		switch(severity){
		case LOG_INFO:
			log.info("VMWare host: "+hostHandle[0]+": " +message);
			break;
		case LOG_WARNING:
			log.warning("VMWare host: "+hostHandle[0]+": " +message);
			break;
		case LOG_CONFIG:
			log.config("VMWare host: "+hostHandle[0]+": " +message);
			break;
		case LOG_SEVERE:
			log.severe("VMWare host: "+hostHandle[0]+": " +message);
			break;
		case LOG_ENTERING:
			//TODO: Implement this?
			break;
		case LOG_EXITING:
			//TODO: Implement this?
			break;
		}
	}
	
	/**
	 * Getter for the host handle of the VMware server (only has meaning for vix (and log files)).
	 * 
	 * @return Vix host handle for this VMware server.
	 */
	protected int getHostHandle(){
		return hostHandle[0];
	}
	
	/**
	 * Getter for the hostname of this VMware server.
	 * 
	 * @return hostname of VMware server.
	 */
	protected String getHost(){
		return hostname;
	}
	
	/*
	 * Log levels
	 */
	private static final int LOG_INFO = 0;
	private static final int LOG_WARNING = 1;
	private static final int LOG_CONFIG = 2;
	private static final int LOG_SEVERE = 3;
	private static final int LOG_ENTERING = 4;
	private static final int LOG_EXITING = 5;
	
	
	
	
	/**
	 * Contains code used for testing VMwareServer and GuestOS.
	 * 
	 * @param args
	 * @throws Exception not handled.
	 */
	public static void main(String[] args) throws Exception{
		String vmxPath = "/var/lib/vmware-server/Virtual Machines/winxp_default/Windows XP Professional.vmx";
        String hostName = "dhcp208-210.ed.ntnu.no";
        int portNumber = 902;
        
    	VMwareServer vws = new VMwareServer(hostName, portNumber);
        vws.connect("joeranvl", "feitfrank");
        
        GuestOS winxp1 = vws.openVM(vmxPath, "thomas", "thomas");
        System.out.println("Power on. Wait for tools, timeout.");
        winxp1.powerOn(true, 30);
        System.out.println("Done waiting for tools.");
        
        winxp1.login();
        //winxp1.takeSnapshot("Default 1", "ClamWin installed, automatic updates disabled.", false);
        //winxp1.loadSnapshot();
       

        //winxp1.revertToSnapshot();
                
        /* ZZZ
        try{
       	int sleepSecs = 15;
        	System.out.println("ZZZ "+sleepSecs+" secs.");
        	Thread.sleep(sleepSecs*1000);
        }catch(InterruptedException e){}
        */
        
        /*
        winxp1.copyFileToGuest("c:\\gah\\foo.txt", "c:\\foo.txt");
        winxp1.copyFileFromGuest("c:\\foo.txt", "c:\\gah\\bar.txt");
        */
        //winxp1.copyFileToGuest("c:\\eicar.com", "c:\\eicar.com");
        //winxp1.runProgramInGuest("c:\\eicar.com", "> c:\\eicarres.txt");
        //winxp1.copyFileFromGuest("c:\\eicarres.txt", "c:\\gah\\eicarres.txt");
        
        //winxp1.powerOff();
        
        vws.disconnect();
	}

}
