package edu.ntnu.item.jt.configloader;

import java.io.BufferedReader;
import java.util.Properties;
import java.io.*;

import edu.ntnu.item.jt.system.Frontend;
import edu.ntnu.item.jt.system.errors.AlreadyInitializedException;

/**
 * 
 * Singleton responsible for maintaining a Properties object. 
 * 
 * Provides static access via getters. Contains logic for reading a properties file from disk.
 * 
 * The properties reader is only intended as a means for providing access to constants, and thus 
 * does not provide any setProperty() functions.
 *   
 * @author Jran Vagnby Lillesand
 * @see java.util.Properties
 *
 */
public class PropertiesReader{
	 
	private Properties props;
	private static PropertiesReader instance;
	private static String synch = "";
	
	private static boolean hasDisplayedInitializeWarning = false;
	
	/**
	 * Internal constructor.
	 * 
	 * @param props
	 */
	private PropertiesReader(Properties props){
		this.props = props;
	}
	
	/**
	 * 
	 * Initialize a properties reader object, reading the properties from a given location on the local file system.
	 * 
	 * This function should ideally be called before getInstance() function is called, as a call to getInstance()
	 * will try to load a default properties file, and will load a blank one if the default properties file cannot be found
	 * on disk.
	 * 
	 * @param configFile path to the properties file on the local file system. Note that any backslashes will have to be escaped. For example "c:\powerscan\powerscan.properties" becomes "c:\\powerscan\\powerscan.properties". The file will have to be structured on a format accepted by the java Properties class. 
	 * @return properties reader instance.
	 * @throws AlreadyInitializedException if the properties reader has already been initialized.
	 * @throws IOException if reading from the file failed. In this case, the properties reader may be initialized again, or calls to getInstance() will try to load the default properties file.
	 * @see java.util.Properties
	 */
	public static PropertiesReader initialize(String configFile) throws AlreadyInitializedException, IOException{
		synchronized(synch){
			if(instance!=null){
				throw new AlreadyInitializedException();
			}
				
			
			Properties props = new Properties();
			props.load(new BufferedReader(new InputStreamReader(new FileInputStream(new File(configFile)))));
			instance = new PropertiesReader(props);
			if(Frontend.debug) System.out.println("initialized properties reader for file: "+configFile+". Can read file: "+new File(configFile).canRead());
			return instance;
		}
		
	}
	
	/**
	 * Returns the properties reader instance.
	 * 
	 * If the properties reader has not been initialized, the default properties file is attempted loaded.
	 * If the default properties file cannot be loaded, an empty Properties object is used, so that the default values will be used for every constant.
	 * 
	 * @return properties reader instance.
	 */
	public static PropertiesReader getInstance(){
		synchronized(synch){
			if(instance != null)
				return instance;
			
			//If properties is not initialized, initialize with standard config name
			try{
				if(!hasDisplayedInitializeWarning) Frontend.getInstance().displayWarning("Using unintialized properties reader, attempting to load default file from "+ "config"+System.getProperty("file.separator")+"powerscan.properties.");
				instance = initialize("config"+System.getProperty("file.separator")+"powerscan.properties");
				
			}catch(IOException e){
				// Nothing to do, return empty Properties.
				if(!hasDisplayedInitializeWarning){
					hasDisplayedInitializeWarning = true;
					Frontend.getInstance().displayError("Unable to load default properties file. Default values will be used.");
				}
				
				return new PropertiesReader(new Properties());
			}catch(AlreadyInitializedException e){
				// Should not happen.
			}

			return instance;
		}		
	}
	
	/**
	 * Reads the value stored with a given key and attempts to cast it to int.
	 * 
	 * If reading the value or casting fails, the default value is returned.
	 * 
	 * @param key key from which to read the value.
	 * @param defaultValue default value to be returned if reading fails.
	 * @return value associated with key or default value.
	 */
	public int getInt(String key, int defaultValue){
		try{
			String stringResult = props.getProperty(key);
			if(stringResult == null){
				Frontend.getInstance().displayWarning("Attempted to read constant with key " +key+", but it could not be read. Default value "+defaultValue+" returned.");
			}
			int result = Integer.parseInt(stringResult);
			return result;
		}catch(Exception e){
			return defaultValue;
		}	
	}
	
	/**
	 * Reads the value stored with a given key and attempts to cast it to double.
	 * 
	 * If reading the value or casting fails, the default value is returned.
	 * 
	 * @param key key from which to read the value.
	 * @param defaultValue default value to be returned if reading fails.
	 * @return value associated with key or default value.
	 */
	public double getDouble(String key, double defaultValue){
		try{
			String stringResult = props.getProperty(key);
			if(stringResult == null){
				Frontend.getInstance().displayWarning("Attempted to read constant with key " +key+", but it could not be read. Default value "+defaultValue+" returned.");
			}
			double result = Double.parseDouble(stringResult);
			return result;
		}catch(Exception e){
			return defaultValue;
		}	
	}
	
	/**
	 * Reads the value stored with a given key and attempts to cast it to boolean.
	 * 
	 * If reading the value or casting fails, the default value is returned.
	 * 
	 * @param key key from which to read the value.
	 * @param defaultValue default value to be returned if reading fails.
	 * @return value associated with key or default value.
	 */
	public boolean getBool(String key, boolean defaultValue){
		try{
			String stringResult = props.getProperty(key);
			if(stringResult == null){
				Frontend.getInstance().displayWarning("Attempted to read constant with key " +key+", but it could not be read. Default value "+defaultValue+" returned.");
			}
			boolean result = Boolean.parseBoolean(stringResult);
			return result;
		}catch(Exception e){
			return defaultValue;
		}	
	}
	
	/**
	 * Attempts to read the string associated with a given key. 
	 * 
	 * Returns a default value if the value cannot be read. 
	 * 
	 * @param key key from which to read the value.
	 * @param defaultValue default value to be returned if reading fails.
	 * @return value associated with key or default value.
	 */
	public String getString(String key, String defaultValue){
		String result = props.getProperty(key);
		if(result != null){
			return result;
		}else{
			Frontend.getInstance().displayWarning("Attempted to read constant with key " +key+", but it could not be read. Default value "+defaultValue+" returned.");
			return defaultValue;
		}
	}
	
	/**
	 * Contains simple test reads.
	 * 
	 * @param args
	 * @throws Exception
	 */
	public static void main(String[] args) throws Exception{
		PropertiesReader.initialize("config\\powerscan.properties");
		
		System.out.println(PropertiesReader.getInstance().getInt("wmware.tools.timeout", 60));
		System.out.println(PropertiesReader.getInstance().getString("scanner.executionpath", "f00"));
		
		System.out.println(PropertiesReader.getInstance().getString("polling.interval.minor", "bar"));
		System.out.println(PropertiesReader.getInstance().getDouble("polling.interval.major", 0.125));
		System.out.println(PropertiesReader.getInstance().getDouble("polling.interval.minor", 0.125));
		
		System.out.println(edu.ntnu.item.jt.configloader.PropertiesReader.getInstance().getString("log.vmware.guestos", "logs"+System.getProperty("file.separator")+"vmware.log"));
	}
	
}
