package edu.ntnu.item.jt.configloader;

import java.io.*;

import edu.ntnu.item.jt.system.errors.XMLValidationException;

import javax.xml.transform.*;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;


import edu.ntnu.item.jt.configloader.XMLErrorHandler;

//imports for the XSD validation
import java.io.File;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLReaderFactory;

/** 
 * A class that contains methods for validating an XML file against a DTD or a W3C XML Schema (XSD).
 * 
 * <p>Technique borrowed from <a href="http://www.roseindia.net/xml/dom/DOMValidateDTD.shtml" target="_blank">http://www.roseindia.net/xml/dom/DOMValidateDTD.shtml</a>.</p>
 * 
 * @author Thomas Langerud
 */
public class XMLValidator {
  
	private final int TEMPDATA_CONSOLE = 1;
	private final int TEMPDATA_FILE = 2;
	
    String parserClass = "org.apache.xerces.parsers.SAXParser";
    String validationFeature = "http://xml.org/sax/features/validation";
    String schemaFeature = "http://apache.org/xml/features/validation/schema";
    String schemaLocationFeature = "http://apache.org/xml/properties/schema/external-noNamespaceSchemaLocation";
    String javaEncoding = "http://apache.org/xml/features/allow-java-encodings";
	
	// controls the printing of debug messages to console
	private final boolean debug = false;
	
	/**
	 * Method that checks an XML file against a DTD file.
	 * If the XML file is well formed and valid according to the DTD, nothing happens
	 * If the XML file is not well formed, or not valid, an exception is raised.
	 * 
	 * Well formed means that the XML file conforms to the rules of the XML language 
	 * (start and end tags, correct nesting, attributes in quotes)
	 * Valid documents are always well formed, but well formed documents need not be valid
	 * 
	 * @param XMLFile Path to the XML file that is to be validated
	 * @param DTDFile Path to the DTD file against which XMLFile is to be validated
	 * @param tempdata value 1 if the XML document should be written to std.out, 2 if a temp file should be used (temp.tmp)
	 */
	public void validateDTD(String XMLFile, String DTDFile, int tempdata) throws XMLValidationException{
		
		/**
		 * If the XML is not well formed, the document can not be parsed.
		 * This exception is catched locally in this class.
		 * If the XML is well formed but not valid according to the given DTF, the error
		 * is sent to the attached XMLErrorHandler which can act according to the severity
		 * of the error.
		 */
		try{
		
			DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
			// force the document parser to also validate the structure
			factory.setValidating(true);
			DocumentBuilder builder = factory.newDocumentBuilder();
			// attach an error handler object to deal with the less severe errors
			builder.setErrorHandler( new XMLErrorHandler() );
			// create a wrapper object for the input XML file
			Document XMLDocument = builder.parse(new FileInputStream(XMLFile));
			DOMSource source = new DOMSource(XMLDocument);
			
			// create the output channels for the XML object
			StreamResult resultStream;
			// create a wrapper for the temp file
			
			File file = File.createTempFile("temp", "tmp");
			file.deleteOnExit();
			FileOutputStream fos = new FileOutputStream(file);
			
			
			switch(tempdata){
			case TEMPDATA_CONSOLE:{
				resultStream = new StreamResult(System.out);
				break;
				}
			case TEMPDATA_FILE:{
				resultStream = new StreamResult(fos);
				break;
				}
			default:{
				// write to console as default
				resultStream = new StreamResult(System.out);
				}
			}
						
			TransformerFactory tf = TransformerFactory.newInstance();
			Transformer transformer = tf.newTransformer();
			transformer.setOutputProperty(OutputKeys.DOCTYPE_SYSTEM, DTDFile);
			transformer.transform(source, resultStream);
			
			// close the file writer object
			fos.close();
			
		}
		catch(Exception e){
			// This occurs if the XML document is not well formed
			throw new XMLValidationException(e.getMessage());
		}	
	}

	/**
	 * Method to validate an XML file against a W3C XML Schema Definition (XSD)
	 * 
	 * @param XMLFile The XML document to validate
	 * @throws XMLValidationException If the XML document is not well formed
	 * @throws FileNotFoundException If the XML document is not found
	 */
	public void validateXSD(String XMLFile) throws XMLValidationException, FileNotFoundException{
	      
		try {
	         XMLReader reader = XMLReaderFactory.createXMLReader(parserClass);
	         reader.setFeature(validationFeature,true);
	         reader.setFeature(schemaFeature,true);
	         reader.setFeature(javaEncoding, true);
	         reader.setErrorHandler(new XMLErrorHandler());
	         reader.setProperty(schemaLocationFeature, PropertiesReader.getInstance().getString("xml.xsd.path", "config.xsd"));
	         reader.parse(XMLFile);
	      } catch (SAXException e) {
	         throw new XMLValidationException(e.getMessage()); 
	      } catch (IOException e) {
	         throw new FileNotFoundException(e.getMessage()); 
	      }
	}
}