package edu.ntnu.item.jt.system;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;

import edu.ntnu.item.jt.system.configbeans.*;
import edu.ntnu.item.jt.system.errors.*;
import java.util.ArrayList;
import java.util.Date;
import java.util.logging.FileHandler;
import java.util.logging.Formatter;
import java.util.logging.Level;
import java.util.logging.LogRecord;
import java.util.logging.Logger;

import edu.ntnu.item.jt.VMware.*;
import edu.ntnu.item.jt.VMware.errors.*;
import edu.ntnu.item.jt.configloader.PropertiesReader;
import edu.ntnu.item.jt.configloader.XMLParser;


/**
 * Responsible for instantiating useful objects based on configbeans created
 * by the PowerScan XMLParser. 
 * 
 * @author Jran Vagnby Lillesand
 * @see XMLParser
 * @see edu.ntnu.item.jt.system.configbeans
 */
public class Connector {
	
	private ArrayList<VMwareServer> vmWareHosts = new ArrayList<VMwareServer>();
	private ArrayList<GuestOS> guestOS = new ArrayList<GuestOS>();
	private ArrayList<Scanner> scanners = new ArrayList<Scanner>();
	private ArrayList<Executor> executors = new ArrayList<Executor>();
	
	private static Logger log = PowerScan.getLogger(PropertiesReader.getInstance().getString("log.system.scanner", "logs"+System.getProperty("file.separator")+"scanner.log"), "edu.ntnu.item.jt.system.scanner");
	
	/**
	 * Constructor for the connector class. Accepts a XML file as input, parses the data using the XMLParser class and
	 * instantiates the appropriate classes. Requires the XMLParser to output configuration beans.
	 * 
	 * Note that this constructor also performs several operations on the instantiated classes:
	 * <ul>
	 * 	<li> VMWare servers are connected to. </li>
	 * 	<li> Virtual Machines (guest os') are loaded, powered on and logged in to. </li>
	 * 	<li> If no snapshot exists for a given VM in advance, one is taken. </li>
	 * 	<li> Scanner objects are instantiated and have their relevant parameters set. </li>
	 * 	<li> Filters are added to scanners </li>
	 * </ul>
	 * 
	 * @param xmlConfig file containing the XML structured PowerScan config.
	 * @throws InvalidFileException if the config file can not be read.
	 * @throws ValidationException if the config file fails DTD vaildation.
	 * @throws ParserException if parsing of the XML file fails.
	 * 
	 * @see XMLParser
	 * @see edu.ntnu.item.jt.system.configbeans
	 */
	public Connector(File xmlConfig) throws InvalidFileException, ParserException{
		
		XMLParser parser = new XMLParser();
		
		Frontend.getInstance().displayInfo("Validating XML config");
		PowerScanBean powerscan = parser.parse(xmlConfig);
		
		log.info("Read "+powerscan.getVMWareHosts().size()+" VMWare hosts from config file. Trying to connect.");
		// Establish connection to VMWareServers
		for(VMwareHostBean host: powerscan.getVMWareHosts()){
			log.info("Trying to establish connection to VMWare host "+host.getHost()+".");
			
			VMwareServer vmWareServer = new VMwareServer(host.getHost(), host.getPortnumber());
			
			try{
				vmWareServer.connect(host.getUsername(), host.getPassword());
				log.info("Successfully connected to VMWare host.");
				Frontend.getInstance().displayInfo("Connected to VMware Host "+ host.getHost());
				vmWareHosts.add(vmWareServer);
				//Connection successful, try to open VMs
				for(VMBean vmBean: host.getVirtualMachines()){
					//Connect to all virtual machines defined for the given VMWare host
					log.info("Attempting to connect to Virtual Machine on path "+vmBean.getVmxPath()+" of VMWare host "+host.getHost());
					try{
						GuestOS guestOS = vmWareServer.openVM(vmBean.getVmxPath(), vmBean.getUsername(), vmBean.getPassword());
						log.info("Successfully connected to virtual machine.");
						
						guestOS.powerOn(true, PropertiesReader.getInstance().getInt("vmware.tools.timeout", 40));
						
						guestOS.login();
						
						log.info("Successfully powered on and logged in to virtual machine.");
						
						// Determine whether the virtual machine is to be used as a scanner or malware executor (or both).
						if(vmBean.getAnalysisTools() != null && vmBean.getAvEngine() != null){
							// Listed as both in XML, display warning in case this was unintentional by user.
							Frontend.getInstance().displayWarning("Virtual machine running from VMX "+vmBean.getVmxPath()+" on host "+host.getHost()+" is running both dynamic analysis tool and anti-virus software.");
							log.warning("Virtual machine running from VMX "+vmBean.getVmxPath()+" on host "+host.getHost()+" is running both dynamic analysis tool and anti-virus software.");
						}
						
						if(vmBean.getAvEngine() != null){
							//Create scan filter
							LogFilter scanFilter;
							if(vmBean.getAvEngine().getFilter() == null){
								// Not log filter element included in XML, create empty scanfilter (which will present the whole log file as result)
								scanFilter = new LogFilter();
							}else{
								scanFilter = new LogFilter(vmBean.getAvEngine().getFilter().getResultIdentifier(), vmBean.getAvEngine().getFilter().getResultPrefix(), vmBean.getAvEngine().getFilter().getResultSuffix());
							}
							
							Scanner scanner = new Scanner(guestOS, vmBean.getAvEngine().getName(),vmBean.getAvEngine().getExecutablePath(), vmBean.getAvEngine().getParameters(), vmBean.getAvEngine().getRemoteLogFilePath(), scanFilter, vmBean.getAvEngine().getAvUpdate(), vmBean.getAvEngine().getRealtimeScanBean());
							scanners.add(scanner);
							log.info("Created scanner for virtual machine at host " +host.getHost()+ " on path "+vmBean.getVmxPath());
						}
						
						if(vmBean.getAnalysisTools() != null && !vmBean.getAnalysisTools().isEmpty()){
							Executor executor = new Executor(vmBean.getAnalysisTools(), guestOS);
							executors.add(executor);
							log.info("Created executor for virtual machine at host "+host.getHost()+ " on path "+vmBean.getVmxPath());
						}
						
						
					}catch(OpenVMException e){
						Frontend.getInstance().displayWarning("Unable to open Virtual Machine on path "+vmBean.getVmxPath()+" of VMWare host "+host.getHost()+". Error message: "+e.getMessage()+".");
						log.warning("Unable to open Virtual Machine on path "+vmBean.getVmxPath()+" of VMWare host "+host.getHost()+". Error message: "+e.getMessage()+". Error code: "+e.getErrCode());
					}catch(LoginFailureException e){
						Frontend.getInstance().displayWarning("Unable to login to Virtual Machine on path "+vmBean.getVmxPath()+" of VMWare host "+host.getHost()+" using username " +vmBean.getUsername()+" and supplied password. Error message: "+e.getMessage()+".");
						log.warning("Unable to login to Virtual Machine on path "+vmBean.getVmxPath()+" of VMWare host "+host.getHost()+" using username " +vmBean.getUsername()+" and supplied password. Error message: "+e.getMessage()+". Error code: "+e.getErrCode());
					}catch(PowerUpFailureException e){
						Frontend.getInstance().displayWarning("Unable to power on Virtual Machine on path "+vmBean.getVmxPath()+" of VMWare host "+host.getHost()+". Error message: "+e.getMessage()+".");
						log.warning("Unable to power on Virtual Machine on path "+vmBean.getVmxPath()+" of VMWare host "+host.getHost()+". Error message: "+e.getMessage()+". Error code: "+e.getErrCode());
					}
				}
			}catch(ConnectionFailedException e){
				if(Frontend.verbose) System.err.println("Warning, unable to connect to VMWare host at "+host.getHost()+". Proceeding without given host.");
				log.warning("Connection to VMWare host "+host.getHost()+ " failed: "+e.getMessage());
				Frontend.getInstance().displayWarning("unable to connect to VMWare host at "+host.getHost()+". Proceeding without given host.");
				
			}
		}
	}

	/**
	 * Getter for the VMwareServer array list.
	 * 
	 * Note that the VMWareServers are connected and ready for use.
	 * 
	 * @return list of VMWare servers instantiated following the parsing of the config file. 
	 */
	public ArrayList<VMwareServer> getVmWareHosts() {
		return vmWareHosts;
	}

	/**
	 * Getter for the Guest OS (virtual machine) array list.
	 * 
	 * Note that the guest OS' are powered on, logged in to and ready for use.
	 * 
	 * @return list of guest OS' instantiated following the parsing of the config file.
	 */
	public ArrayList<GuestOS> getGuestOS() {
		return guestOS;
	}

	
	/**
	 * Getter for the scanner array list.
	 * 
	 * Note that the scanners are associated with the necessary parameters (including guest OS) and ready for use.
	 * 
	 * @return list of scanners instantiated following parsing of the config file.
	 */
	public ArrayList<Scanner> getScanners() {
		return scanners;
	}
	
	public ArrayList<Executor> getExecutors(){
		return executors;
	}
}
