/*
 * Created on 1. april. 2008
 *
 */
package edu.ntnu.item.jt.system;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.logging.FileHandler;
import java.util.logging.Formatter;
import java.util.logging.Level;
import java.util.logging.LogRecord;
import java.util.logging.Logger;

import javax.swing.ProgressMonitor;

//import edu.ntnu.item.jt.VMware.VMwareServer;
import edu.ntnu.item.jt.VMware.GuestOS;
import edu.ntnu.item.jt.VMware.errors.CreateSnapshotException;
import edu.ntnu.item.jt.VMware.errors.FileCopyException;
import edu.ntnu.item.jt.VMware.errors.FileExecutionException;
import edu.ntnu.item.jt.VMware.errors.LoginFailureException;
import edu.ntnu.item.jt.VMware.errors.NoSnapshotAvailableException;
import edu.ntnu.item.jt.VMware.errors.NoSnapshotLoadedException;
import edu.ntnu.item.jt.VMware.errors.RevertToSnapshotFailedException;
import edu.ntnu.item.jt.configloader.PropertiesReader;
import edu.ntnu.item.jt.system.configbeans.*;
import edu.ntnu.item.jt.system.errors.InvalidFileException;

/**
 * This class contains all tools that is installed on a given Guest OS
 * 
 * @author Thomas Langerud
 *
 */
public class Executor{

	private ArrayList<DynamicAnalysisToolBean> tools;
	private GuestOS guestOS;
	private String guestOSid;
	private boolean executeSample = false;
	private ProgressMonitor progressMonitor;
	
	// Instantiate 
	private static Logger log = PowerScan.getLogger(PropertiesReader.getInstance().getString("log.system.executor", "logs"+System.getProperty("file.separator")+"executor.log"), "edu.ntnu.item.jt.system.executor");
		
	
	/**
	 * Private logger to assist in parsing log messages .
	 * 
	 * @param severity severity level of the occurence.
	 * @param message log message.
	 */
	private void log(int severity, String message){
		
		switch(severity){
		case LOG_INFO:
			log.info("Executor '"+ guestOSid +"': " +message);
			break;
		case LOG_WARNING:
			log.warning("Executor '"+ guestOSid +"': " +message);
			break;
		case LOG_CONFIG:
			log.config("Executor '"+ guestOSid +"': " +message);
			break;
		case LOG_SEVERE:
			log.severe("Executor '"+ guestOSid +"': " +message);
			break;
		case LOG_ENTERING:
			//TODO: Implement this?
			break;
		case LOG_EXITING:
			//TODO: Implement this?
			break;
		}
	}
	
	/*
	 * Log levels
	 */
	private static final int LOG_INFO = 0;
	private static final int LOG_WARNING = 1;
	private static final int LOG_CONFIG = 2;
	private static final int LOG_SEVERE = 3;
	private static final int LOG_ENTERING = 4;
	private static final int LOG_EXITING = 5;	
	
	/**
	 * Constructs a new Executor representing a GuestOS
	 * 
	 * @param daToolBeans An ArrayList containing the tools installed on one given Guest OS
	 * @param guestOS A reference to the Guest OS on which the tools are installed
	 */
	public Executor(ArrayList<DynamicAnalysisToolBean> daToolBeans, GuestOS guestOS) {
		log(LOG_INFO, "Creating new executor object for guestOS with handle "+guestOS.getVMHandle()+ " on host "+guestOS.getHost());
		tools = daToolBeans;
		this.guestOS = guestOS;
		guestOSid = guestOS.getHost() + ":" + guestOS.getvmxPath();
		
	}
	
	/**
	 * Executes the tools installed on the given Guest OS. If more than one, the tools are started in sequence. No callback mechanism is used.
	 * 
	 * @param localSamplePath path to the malware sample to be analyzed.
	 * @param takeSnapshotBeforeExecute boolean value to indicate whether a snapshot should be taken before the tools are executed.
	 * @return list of analysis tool results.
	 * @throws InvalidFileException 
	 * @throws CreateSnapshotException
	 * @throws FileCopyException
	 * @throws FileExecutionException
	 */
	public ArrayList<DynamicAnalysisResult> executeTools(File localSamplePath, boolean takeSnapshotBeforeExecute) throws InvalidFileException,CreateSnapshotException, FileCopyException, FileExecutionException{
		
		ArrayList<DynamicAnalysisResult> resultList = new ArrayList<DynamicAnalysisResult>(); 
		String toolName ="";
		
		try {
			guestOS.login();
		}
		catch(LoginFailureException e) {
			if(Frontend.debug) System.out.println("Login failed on VM " + guestOSid);
		}
		
		// Check that the sample file is actually okay and readable
		if(!localSamplePath.isFile() || !localSamplePath.canRead()){
			log(LOG_WARNING, "Unable to read local file. Aborting malware execution.");
			throw new InvalidFileException("Unable to read malware sample file on local host.");
		}	
		
		//Find filename of file to be scanned		
		String malwareFileName = localSamplePath.getAbsolutePath().substring(localSamplePath.getAbsolutePath().lastIndexOf(File.separatorChar)+1);

		// Take snapshot?
		if(takeSnapshotBeforeExecute){
			log(LOG_INFO, "Taking snapshot of virtual machine state on VM " + guestOSid+ ".");
			guestOS.takeSnapshot("Before Scan", "Snapshot automatically taken prior to performing scan.", false);
			log(LOG_INFO, "Snapshot successfully taken on VM " +guestOSid+ ".");
		}		
	
		// Read location to which to copy the file.
		String remoteMalwarePath = PropertiesReader.getInstance().getString("executor.executionpath", "c:\\");

		
		// Copy the sample over to the VM
		try {
			log(LOG_INFO, "Attempting to copy local sample file to "+remoteMalwarePath+malwareFileName+" on guest OS.");
			Frontend.getInstance().displayInfo("Copying sample file to VM "+guestOSid);
			guestOS.copyFileToGuest(localSamplePath.getAbsolutePath(), remoteMalwarePath+malwareFileName);
			log(LOG_INFO, "Successfully copied local sample file to "+remoteMalwarePath+malwareFileName+" on guest OS.");
			Frontend.getInstance().displayInfo("Done copying sample file to VM");
		}
		catch(FileCopyException e) {
			Frontend.getInstance().displayError("Error copying malware sample to remote system " + guestOSid);
			log(LOG_SEVERE, "Error copying malware sample to remote machine "+ guestOSid +". Message: "+e.getMessage());
			throw e;
		}
		
		// Log in with the GUI user
		// Returns error message : insufficient rights in guest OS
		/*
		try {
			log(LOG_INFO,"Attempting login with the GUI console user");
			Frontend.getInstance().displayInfo("Logging in on server console");
			guestOS.logInGuiConsoleUser();
		}
		catch(LoginFailureException e) {
			log(LOG_SEVERE,"Login with the GUI console user failed. Message: "+e.getMessage());
			Frontend.getInstance().displayWarning("Logging in to server with console user is indicated as failed. Started programs may not be visible");
		}
		*/
		
		// Start execution of the all the tools in the VM
		log(LOG_INFO, "Starting execution of tools in daToolBean list");
		for( DynamicAnalysisToolBean daToolBean: tools ) {

			toolName = daToolBean.getName();
			
			try {
				log(LOG_INFO, "Executing tool " + toolName + " from path " + daToolBean.getToolExecutablePath());
				
				if(daToolBean.getExecuteMalwareExplicitly()) {
					executeSample = true;
				}
				
				String parameters = "";
				if(daToolBean.getToolParameters() != null && daToolBean.getToolParameters() != "") {
					if(daToolBean.getToolParameters().contains("$samplePath")) {
						parameters = daToolBean.getToolParameters().replace("$samplePath", remoteMalwarePath+malwareFileName);
					}
					else parameters = daToolBean.getToolParameters();
				}
				
				guestOS.runProgramInGuestNonBlocking(daToolBean.getToolExecutablePath(),parameters);
				Frontend.getInstance().displayInfo("Tool " + toolName + " is now running on VM " + guestOSid+".");	

			}
			catch(FileExecutionException fee) {
				log(LOG_SEVERE,"Unable to execute analysis tool on remote system");
				Frontend.getInstance().displayWarning("Unable to execute analysis tool "+ toolName + " on remote system " + guestOSid);
				//throw fee;
			}
		}
		if(executeSample) {
			// Also execute the malware sample
			try {
				log(LOG_INFO, "Trying to execute the malware sample explicitly");
				guestOS.runProgramInGuestNonBlocking(remoteMalwarePath+malwareFileName, "");
				Frontend.getInstance().displayInfo("Malware sample now executed. ");
				log(LOG_INFO,"Malware sample executed explicitly");
			}
			catch(FileExecutionException e) {
				Frontend.getInstance().displayError("Error executing malware sample on remote system. Please see to log file. ");
				log(LOG_SEVERE,"Unable to execute malware sample on remote system. Message: "+e.getMessage());
				//throw e;
			}
		}
	

		int sleepSec = PropertiesReader.getInstance().getInt("executor.overall.timeout", 30);
		long sleepTime =  sleepSec * 1000;
		
		Frontend.getInstance().displayInfo("All analysis tools have been started on VM "+guestOSid+", user interaction might be required. Log in to VM using username/password " + guestOS.getUsername()+"/" + guestOS.getPassword());
		log(LOG_INFO,"All tools started, now going to sleep for "+sleepTime+" ms." );
		
		Frontend.getInstance().displayInfo("Now sleeping - awaiting user interaction");
		
		// call to Frontend. This call will return after $sleepSec seconds. 
		Frontend.getInstance().sleep(sleepSec);
		
		Frontend.getInstance().displayInfo("Done sleeping - completing task");
		
		log(LOG_INFO,"Done sleeping, now going to fetch result log files" );		
		
		// Start copying over the result files
		for( DynamicAnalysisToolBean daToolBean: tools ) {
			toolName = daToolBean.getName();
			if( (daToolBean.getToolResultFile() != null) && ( !daToolBean.getToolResultFile().equals("") ) ){
			
				// Read name of result file
				File resultFile = new File(daToolBean.getToolResultFile());
				String resultFileName = resultFile.getAbsolutePath().substring(resultFile.getAbsolutePath().lastIndexOf(File.separatorChar)+1);
				
				// Try to create local result directory if it does not already exist
				new File(PropertiesReader.getInstance().getString("executor.localResultPath", "c:\\")).mkdirs();
				
				// Build full path for the local result file
				String localResultPath = PropertiesReader.getInstance().getString("executor.localResultPath", "c:\\")+resultFileName;
				
				try {
					// Copy result file back
					log(LOG_INFO,"Attempting to copy back remote file " + daToolBean.getToolResultFile());
					Frontend.getInstance().displayInfo("Copying result file back for tool " + daToolBean.getName());
					guestOS.copyFileFromGuest(daToolBean.getToolResultFile(), localResultPath);		
					
					log(LOG_INFO,"Successfully copied remote result file  " + daToolBean.getToolResultFile() + " back to local system at "+localResultPath);
					resultList.add(new DynamicAnalysisResult("Tool " +toolName+ " completed, result file is located at " + localResultPath));				
				}
				catch(FileCopyException fce) {
					log(LOG_SEVERE, "Copy of remote result file(s) failed from "+ daToolBean.getToolResultFile() +" to "+localResultPath);
					Frontend.getInstance().displayWarning("Copy of remote result file(s) failed from "+ daToolBean.getToolResultFile());
					resultList.add(new DynamicAnalysisResult("Tool "+toolName+" completed, but system was unable to copy result file back"));
				}
				
			}
			else resultList.add(new DynamicAnalysisResult("Tool "+toolName+" completed. No result file is declared."));
		}
		return resultList;
	}
	
	/**
	 * Revert the current guest OS to the stored snapshot
	 */
	public void revertToSnapshot(){
		try {
			log(LOG_INFO,"Trying to revert VM "+guestOSid+" to snapshot");
			Frontend.getInstance().displayInfo("Reverting VM with handle "+guestOS.getVMHandle()+ " on host "+guestOS.getHost()+" to snapshot");
			guestOS.revertToSnapshot(false);
			log(LOG_INFO,"Revert to snapshot succeded");
		}
		catch(NoSnapshotLoadedException e) {
			log(LOG_WARNING,"Error reverting "+guestOSid+" to snapshot failed, message : "+e.getMessage()+". Error code is "+e.getErrCode());
			Frontend.getInstance().displayError("Revert to snapshot failed; message: "+e.getMessage()); 
		}
		catch(RevertToSnapshotFailedException e) {
			log(LOG_WARNING,"Error reverting "+guestOSid+" to snapshot failed, message : "+e.getMessage()+". Error code is "+e.getErrCode());
			Frontend.getInstance().displayError("Revert to snapshot failed; message: "+e.getMessage());
		}
		catch(NoSnapshotAvailableException e) {
			log(LOG_WARNING,"Error reverting "+guestOSid+" to snapshot failed, message : "+e.getMessage()+".");
			Frontend.getInstance().displayError("Revert to snapshot failed; No snapshot available. Message: "+e.getMessage() );
		}
	}
}
