package edu.ntnu.item.jt.system;

import java.io.*;
import java.util.ArrayList;
import org.apache.commons.cli.*;
import edu.ntnu.item.jt.GUI.GUIFrontend;
import edu.ntnu.item.jt.system.errors.*;

/**
 * Provides a frontend to the application for the user.
 * 
 *  Can be extended to implement new types of UI as they are written (using inheritance and modifying the getInstance() function).
 * 
 * @author Jran Vagnby Lillesand
 *
 */
public class Frontend {
	
	private static Options options; // Description of command line options

	// Command line arguments
	private static final String PROPERTIES_LOCATION_OPTION = "SETTINGS";
	private static final String PROPERTIES_LOCATION_OPTION_SHORT = "s";
	private static final String DEFAULT_PROPERTIES_FILE = "config"+System.getProperty("file.separator")+"powerscan.properties";
	
	private static final String XML_LOCATION_OPTION = "CONFIG";
	private static final String XML_LOCATION_OPTION_SHORT = "c";
	private static final String DEFAULT_XML_CONFIG_FILE = "config"+System.getProperty("file.separator")+"config.xml";
	
	private static final String AV_UPDATE_OPTION = "UPDATE";
	private static final String AV_SCAN_OPTION = "SCAN";
	private static final String AV_EXECUTE_MALWARE = "EXECUTE";
	private static final String DYNAMIC_ANALYSIS = "ANALYZE";
	
	private static final String HELP_USAGE = "java -jar powerscan.jar [arguments].";
	
	public static final int UI_SHELL = 1;
	public static final int UI_GUI = 2;
	private static int ui = UI_SHELL;
	private static Frontend instance;
	
	public static boolean debug = false;
	public static boolean verbose = false;
	
	static{
		
		
		options = new Options();
		
		options.addOption("GUI", false, "Determines whether to use a GUI.");
		
		options.addOption(AV_UPDATE_OPTION, false, "Performs a update of all the AV engines listed in the XML configuration file. If update can not be done for a given scanner, the user will be instructed on how to perform it.");
		options.addOption(AV_SCAN_OPTION, true, "Performs a surface scan of the supplied malware sample with all scan engines configured in the XML config file.");
		options.addOption(AV_EXECUTE_MALWARE, true, "Executes the supplied malware sample on any virtual machine supporting real-time anti-virus detection and extracts results.");
		options.addOption(DYNAMIC_ANALYSIS, true, "Perform dynamic analysis of the supplied sample with all (if any) registered dynamic analysis tools from XML.");
		
		options.addOption(PROPERTIES_LOCATION_OPTION_SHORT, PROPERTIES_LOCATION_OPTION, true, "Properties file location. Default location: "+Frontend.DEFAULT_PROPERTIES_FILE);
		options.addOption(XML_LOCATION_OPTION_SHORT, XML_LOCATION_OPTION, true, "XML Config file location. Default location: "+Frontend.DEFAULT_XML_CONFIG_FILE);
	}
	
	/**
	 * Private constructor.
	 * Edit: changed to protected due to inheritance. Not an optimal solution, should look into using an interface and a singleton factory
	 */
	protected Frontend(){
		
	}
	
	/**
	 * Getter for the Frontend singleton. 
	 * 
	 * Can be extended to support different UIs. 
	 * 
	 * @return instance of the frontend.
	 */
	public static synchronized Frontend getInstance(){
		//TODO: Implement support for other UIs as they are created.
		if(instance == null){
			switch(ui){
			case UI_SHELL:
				instance =  new Frontend();
				break;
			case UI_GUI:
				instance = GUIFrontend.getInstance();
				break;
			default:
				instance =  new Frontend();
			}
			
		}
			
		return instance;
	} 
	
	
	/**
	 * Parses input parameters and starts operation appropriately.
	 * 
	 * @param args command line arguments.
	 */
	public static void main(String[] args){
		
		long startTime = System.currentTimeMillis();
		
		// Parse args according to options
		
		String properties = "";
		String xmlConfig="";
		
		try{
			CommandLineParser parser = new PosixParser();
			
			CommandLine cmd = parser.parse(options, args);
			
			// First, check if GUI is attempted loaded.
			if(cmd.hasOption("GUI")){
				ui = UI_GUI;
				// initialize GUI window
				Frontend.getInstance();
				return;
			}	
			// Try to get and set optional arguments
			if(cmd.hasOption(Frontend.PROPERTIES_LOCATION_OPTION)){
				properties = cmd.getOptionValue(Frontend.PROPERTIES_LOCATION_OPTION);
			}else{
				Frontend.getInstance().displayInfo("No settings file location supplied, using default location: "+Frontend.DEFAULT_PROPERTIES_FILE);
				properties = Frontend.DEFAULT_PROPERTIES_FILE;
			}
			
			// Try to get and set XML Config file
			if(cmd.hasOption(Frontend.XML_LOCATION_OPTION)){
				if(!new File(cmd.getOptionValue(Frontend.XML_LOCATION_OPTION)).canRead()){
					Frontend.getInstance().displayError("Unable to read XML config file "+cmd.getOptionValue(Frontend.XML_LOCATION_OPTION));
					new HelpFormatter().printHelp(Frontend.HELP_USAGE, options);
					System.exit(1);
				}else{
					xmlConfig = cmd.getOptionValue(Frontend.XML_LOCATION_OPTION);
				}
			}else{
				Frontend.getInstance().displayInfo("No XML config location supplied, using default location: "+Frontend.DEFAULT_XML_CONFIG_FILE);
				xmlConfig = Frontend.DEFAULT_XML_CONFIG_FILE;

			}
			
			// Check operational parameters and perform requested operation.
			try{

				if(cmd.hasOption(Frontend.AV_UPDATE_OPTION) || cmd.hasOption(Frontend.AV_SCAN_OPTION) || cmd.hasOption(Frontend.AV_EXECUTE_MALWARE) || cmd.hasOption(Frontend.DYNAMIC_ANALYSIS)){
					PowerScan powerscan = null;
					try {
						powerscan = new PowerScan(xmlConfig, properties);
					}
					catch(XMLValidationException e) {
						Frontend.getInstance().displayError("Error while validating XML config file, error: \n"+e.getMessage());
					}
					catch(FileNotFoundException e) {
						Frontend.getInstance().displayError("Error while validating XML config file, error: \n"+e.getMessage());
					}
					
					if(cmd.hasOption(Frontend.AV_UPDATE_OPTION)){
						//attempt to update AV engines
						ArrayList<AVUpdateResult> avUpdateResults = powerscan.update();
						Frontend.getInstance().displayUpdateResults(avUpdateResults);
					}else if(cmd.hasOption(Frontend.AV_SCAN_OPTION)){
						// If not update, then malware sample should be submitted.
						
						String malwareSample = cmd.getOptionValue(Frontend.AV_SCAN_OPTION);
						if(!new File(malwareSample).canRead()){
							new HelpFormatter().printHelp(Frontend.HELP_USAGE, "Unable to read file "+malwareSample, options, "");
							System.exit(1);
						}else{
							// File is okay, perform scan.
							ArrayList<ScanResult> scanResults = powerscan.scan(malwareSample);
							Frontend.getInstance().displayScanResults(scanResults);
						}
					}else if(cmd.hasOption(Frontend.AV_EXECUTE_MALWARE)){
						String malwareSample = cmd.getOptionValue(Frontend.AV_EXECUTE_MALWARE);
						
						if(!new File(malwareSample).canRead()){
							new HelpFormatter().printHelp(Frontend.HELP_USAGE, "Unable to read file "+malwareSample, options, "");
							System.exit(1);
						}else{
							// File is okay, execute malware.
							ArrayList<ScanResult> scanResults = powerscan.executeMalwareSample(malwareSample);
							Frontend.getInstance().displayScanResults(scanResults);
						}
					}else if(cmd.hasOption(Frontend.DYNAMIC_ANALYSIS)){
						String malwareSample = cmd.getOptionValue(Frontend.DYNAMIC_ANALYSIS);
						
						if(!new File(malwareSample).canRead()){
							new HelpFormatter().printHelp(Frontend.HELP_USAGE, "Unable to read file "+malwareSample, options, "");
							System.exit(1);
						}else{
							ArrayList<DynamicAnalysisResult> analysisResults = powerscan.executeInAnalysisTool(malwareSample);
							Frontend.getInstance().displayAnalysisToolsResult(analysisResults);
						}
					}
				}else{
					new HelpFormatter().printHelp(Frontend.HELP_USAGE, options);
					System.exit(1);
				}
				
				
			}catch(InvalidFileException e){
				Frontend.getInstance().displayError("An error occured while trying to read XML config: "+e.getMessage());
				System.exit(1);
			}catch(ParserException e){
				Frontend.getInstance().displayError("An error occured while trying to parse XML config: "+e.getMessage());
				System.exit(1);
			}
		

		}catch(ParseException e){
			Frontend.getInstance().displayError("Unable to parse command line arguments: "+e.getMessage());
			new HelpFormatter().printHelp(Frontend.HELP_USAGE, options);
		}
		
		//System.out.println("Total execution time: "+ (System.currentTimeMillis()-startTime)/1000);	
	}
	
	/**
	 * Display the result of a update operation to the user. Should be adjusted to the current UI.
	 * 
	 * @param results a list of AV engine update results.
	 */
	public void displayUpdateResults(ArrayList<AVUpdateResult> results){ 
		Frontend.getInstance().displayInfo("");
		Frontend.getInstance().displayInfo("-----------Result of update operation------------");
		
		boolean hit = false;
		
		for(AVUpdateResult result: results){
			if(result.getResultState() == AVUpdateResult.UPDATE_SUCCESSFUL){
				if(!hit){
					Frontend.getInstance().displayInfo("Successfully updated engines:");
					hit = true;
				}
				Frontend.getInstance().displayInfo(result.getAVEngineName());
				for(String successLine: result.getResultString()){
					Frontend.getInstance().displayInfo("	"+successLine);
				}
			}
		}
		
		if(hit)
			Frontend.getInstance().displayInfo("");
		
		hit = false;
		
		for(AVUpdateResult result: results){
			if(result.getResultState() == AVUpdateResult.FULL_LOG_RETURNED || result.getResultState() == AVUpdateResult.SUCCESS_INDICATOR_NOT_FOUND){
				if(!hit){
					Frontend.getInstance().displayInfo("The following update operations returned a log file without indicating success and must have their log files review manually:");
					hit = true;
				}
				Frontend.getInstance().displayInfo(result.getAVEngineName() + " running at "+result.getScanner().getScanMachine().getHost() +" with OS login/pass "+result.getScanner().getScanMachine().getUsername()+"/"+result.getScanner().getScanMachine().getPassword()+":");
				for(String updateLogLine: result.getResultString()){
					Frontend.getInstance().displayInfo("	"+updateLogLine);
				}
			}
		}
		
		if(hit){
			Frontend.getInstance().displayInfo("Note that snapshots have been taken (unless otherwise has been noted during execution).");
			Frontend.getInstance().displayInfo("Please log in to any unsuccessfully updated engines, perform manual update and take snapshot.");
			Frontend.getInstance().displayInfo("");
		}
		
		hit = false;
		
		for(AVUpdateResult result: results){
			if(result.getResultState() == AVUpdateResult.LOG_NOT_SUPPORTED || result.getResultState() == AVUpdateResult.NO_LOG_FOUND){
				if(!hit){
					Frontend.getInstance().displayInfo("Engines that does not provide verbose update log and should be reviewed manually:");
					hit = true;					
				}
				
				Frontend.getInstance().displayInfo(result.getAVEngineName() + " running at "+result.getScanner().getScanMachine().getHost() +" with OS login/pass "+result.getScanner().getScanMachine().getUsername()+"/"+result.getScanner().getScanMachine().getPassword()+".");
			}	
		}
		
		if(hit){
			Frontend.getInstance().displayInfo("Please login to these virtual machines with the provided usernames/passwords and assure that they are updated and ready to perform scans. Remember to take snapshots if any further changes are made.");
			Frontend.getInstance().displayInfo("");
		}
		hit = false;
		
		for(AVUpdateResult result: results){
			if(result.getResultState() == AVUpdateResult.UPDATE_NOT_SUPPORTED){
				if(!hit){
					Frontend.getInstance().displayInfo("The following engines does not support automatic updates and must be updated manually:");
					hit = true;					
				}
				
				Frontend.getInstance().displayInfo(result.getAVEngineName() + " running at "+result.getScanner().getScanMachine().getHost() +" with OS login/pass "+result.getScanner().getScanMachine().getUsername()+"/"+result.getScanner().getScanMachine().getPassword()+".");
			}	
		}
		
		if(hit){
			Frontend.getInstance().displayInfo("Please login to these virtual machines with the provided usernames/passwords and perform manual update."); 
			Frontend.getInstance().displayInfo("Remember to take snapshot after updating.");
			Frontend.getInstance().displayInfo("");
		}
		hit = false;
		
		
		for(AVUpdateResult result: results){
			if(result.getResultState() == AVUpdateResult.OPERATION_TIMEOUT){
				if(!hit){
					Frontend.getInstance().displayInfo("The following engines timed out due to global timeout:");
					hit = true;					
				}
				
				Frontend.getInstance().displayInfo(result.getAVEngineName() + " running at "+result.getScanner().getScanMachine().getHost() +" with OS login/pass "+result.getScanner().getScanMachine().getUsername()+"/"+result.getScanner().getScanMachine().getPassword()+".");
			}	
		}
		
		if(hit){
			Frontend.getInstance().displayInfo("Either increase the timeout and try again or log in to these virtual machines and perform updates manually."); 
			Frontend.getInstance().displayInfo("Remember to take snapshot after performing manual updates.");
			Frontend.getInstance().displayInfo("");
		}
		hit = false;
		
		for(AVUpdateResult result: results){
			if(result.getResultState() == AVUpdateResult.NOT_INITIATED){
				if(!hit){
					Frontend.getInstance().displayInfo("The following engines did not initiate update:");
					hit = true;					
				}
				
				Frontend.getInstance().displayInfo(result.getAVEngineName() + " running at "+result.getScanner().getScanMachine().getHost() +" with OS login/pass "+result.getScanner().getScanMachine().getUsername()+"/"+result.getScanner().getScanMachine().getPassword()+".");
			}	
		}
		
		if(hit){
			Frontend.getInstance().displayInfo("Please review these engines manually."); 
			Frontend.getInstance().displayInfo("");
		}
		hit = false;
	}
	
	/**
	 * Displays the result of a scan to the user. The presentation should be done in a suitable manner for the
	 * running UI.
	 * 
	 * @param results list of scan results to be presented.
	 */
	public void displayScanResults(ArrayList<ScanResult> results){
		// Try to print date of last update
		try{
			Frontend.getInstance().displayInfo("Last virus definition update operation started from this client: "+PowerScan.getLastUpdate()+".");
		}catch(NoMatchFoundException e){
			Frontend.getInstance().displayWarning("Virus definitions have never been updated from this client.");
		}
		
		Frontend.getInstance().displayInfo("-----------Result of scan------------");
	
		for(ScanResult result: results){
			Frontend.getInstance().displayInfo(result.resultSummary());
		}
		
		Frontend.getInstance().displayInfo("-----------End of result ------------");
	}
	
	public void displayAnalysisToolsResult(ArrayList<DynamicAnalysisResult> results) {
		Frontend.getInstance().displayInfo("---------Result of analysis tool execution--------");
		
		for(DynamicAnalysisResult result : results) {
			Frontend.getInstance().displayInfo(result.getAnalysisResult() );
		}
		Frontend.getInstance().displayInfo("---------------End of result----------------------");
	}
	
	/**
	 * Display information regarding the system.
	 * 
	 * @param message the message to be displayed.
	 */
	public void displayInfo(String message){
		System.out.println(message);
	}
	
	/**
	 * Display warnings regarding the system.
	 * 
	 * @param message the message to be displayed.
	 */
	public void displayWarning(String message){
		System.err.println("! WARNING ! "+message);
	}

	/**
	 * Display errors regarding the system.
	 *  
	 * @param message the message to be displayed.
	 */
	public void displayError(String message){
		System.err.println("!!! ERROR !!! "+message);
	}
	
	/**
	 * This method will sleep for a given amount of seconds, and update the UI every second.
	 * @param sleepSec The number of seconds to sleep in total.
	 */
	public void sleep(int sleepSec) {
		Frontend.getInstance().displayInfo("Waiting "+sleepSec+" seconds for user interaction. Press the enter key to skip.");
		BufferedReader keyboard = new BufferedReader(new InputStreamReader(System.in));
		
		// Set polling interval
		double pollingInterval = 0.259;
		
		for(int i = 0; i < sleepSec/pollingInterval; i++){
			
			try{
				Thread.sleep((long ) (pollingInterval*1000));
			}catch(InterruptedException e){
				// Should not happen, but if it does, ignore it.
			}
			
			try{
				// Check if enter has been pressed
				if(keyboard.ready()){
					// Break for-loop
					i = (int) (sleepSec/pollingInterval);
					// Reset buffer some that the function works the next time
					while(keyboard.ready()) keyboard.readLine();
				}else{
					if((pollingInterval*i) % 10 < pollingInterval && i != 0){
						Frontend.getInstance().displayInfo("Slept "+(int) (pollingInterval*i)+" seconds of "+sleepSec+" total.");
					}	
				}
			}catch(IOException e){
				// Nothing to be done if this happens
			}
		}
	}

}
