package edu.ntnu.item.jt.system;

import java.util.ArrayList;
import edu.ntnu.item.jt.system.errors.*;
import java.io.*;

/**
 * A representation of the output format of a given antivirus engine.
 * 
 * Keeps record of ways to interpret, analyze and filter an AVE's scan output.
 * 
 * @author Jran Vagnby Lillesand
 *
 * @see Scanner
 */
public class LogFilter {
	
	private String resultIdentifier;
	private String resultPrefix;
	private String resultSuffix;
	
	/**
	 * Constructor for a log filter that looks for a certain identifier string in each line and extract the information between
	 * a given prefix and suffix in that string.
	 * 
	 * @param resultIdentifier string that identifies a line containing a match.
	 * @param resultPrefix unique string (for the line containing the match) identifying the beginning of the result (i.e. the actual match). If the parameter is null, it will be assumed the result is at the beginning of the line.  
	 * @param resultSuffix unique string (for the line containing the match) identifying the end of the result (i.e. the actual match). If the parameter is null, it will be assumed the result is at the end of the line.
	 */
	public LogFilter(String resultIdentifier, String resultPrefix, String resultSuffix) {
		this.resultIdentifier = resultIdentifier;
		this.resultPrefix = resultPrefix;
		this.resultSuffix = resultSuffix;
	}
	
	/**
	 * Constructor for a log filter that looks for a certain identifier string in each line and returns every line containing it.
	 * 
	 * @param resultIdentifier string that identifies a line containing a match.
	 */
	public LogFilter(String resultIdentifier){
		this.resultIdentifier = resultIdentifier;
	}
	
	/**
	 * Constructor for a log filter that returns every line of the log file.
	 */
	public LogFilter(){
		
	}
	
	/**
	 * Reads a log file and extracts parts of it based on the prefixes and suffixes associated with this instance.
	 * 
	 * If no prefix and suffix is defined, every line containing the resultIdentifier set in the constructor is returned.
	 * 
	 * Uses simple String matching to identify the scan result in the log file. Supports multiple matches per log.
	 * 
	 * @param log the log file to be filtered.
	 * @return a generic representation of the scan result.
	 * @throws NoMatchFoundException no match was found during the scan, i.e. the scanned file was reported clean. 
	 * @throws FileNotFoundException the input file to be parsed could not be found.
	 * @throws IOException an IO exception occured while reading the file before any result was found (if the exception is thrown after a result is found, it is ignored).
	 */
	public ArrayList<String> filter(File log) throws NoMatchFoundException, FileNotFoundException, IOException{
		BufferedReader br = new BufferedReader(new InputStreamReader(new FileInputStream(log)));
		ArrayList<String> result = new ArrayList<String>();
		
		try{
			while(br.ready()){
				String line = br.readLine();
				
				if(resultIdentifier == null){
					result.add(line);
				}else if(line.contains(resultIdentifier)){
					int resStartPos;
					int resEndPos;
					
					if(!(resultPrefix == null || resultPrefix.equals(""))){
						resStartPos=line.indexOf(resultPrefix)+resultPrefix.length();
					}else{
						// No resultPrefix signifies match is at beginning of line.
						resStartPos = 0;
					}
						
					
					if(!(resultSuffix == null || resultSuffix.equals(""))){
						resEndPos=line.indexOf(resultSuffix);
					}else{
						// No resultSuffix signifies match is at end of line.
						resEndPos = line.length();
					}
						
					result.add(line.substring(resStartPos, resEndPos));
				}
				
			}
			
			br.close();
		}catch(IOException e){
			Frontend.getInstance().displayWarning("IO Error occured while reading log file "+log.getAbsolutePath());
			
			// If there is some result, allow it to be returned
			if(result.isEmpty())
				throw new IOException();
		}
		
		if(result.isEmpty())
			throw new NoMatchFoundException();
		
		return result;	
	}
	
}
