package edu.ntnu.item.jt.system;

import edu.ntnu.item.jt.VMware.errors.*;
import edu.ntnu.item.jt.VMware.*;
import edu.ntnu.item.jt.configloader.PropertiesReader;
import edu.ntnu.item.jt.system.errors.*;
import java.io.*;
import java.util.ArrayList;
import java.util.logging.FileHandler;
import java.util.logging.Formatter;
import java.util.logging.Level;
import java.util.logging.LogRecord;
import java.util.logging.Logger;
import java.util.prefs.Preferences;
import java.util.Date;
import java.text.SimpleDateFormat;

/**
 * Central "control point" for the PowerScan application. Provides some simple functionality between the
 * UI and the executing classes.
 * 
 * @author Jran Vagnby Lillesand
 *
 */
public class PowerScan {
	
	private Connector connector;
	
	
	/**
	 * Instantiates a new PowerScan object.
	 * 
	 * Accepts the system's two configuration files (xml config and properties) and initalizes/parses these.
	 * The instantiated PowerScan object should be used to perform PowerScan system operations.
	 * 
	 * @param xmlConfigFile path to the XML config file with the setup parameters for the system.
	 * @param properties path to the properties file containing the system constants.
	 * @throws ParserException if parsing of the XML config file failed.
	 * @throws InvalidFileException if the XML config file could not be read.
	 */
	public PowerScan(String xmlConfigFile, String properties) throws ParserException, InvalidFileException, XMLValidationException, FileNotFoundException{
		try{
			PropertiesReader.initialize(properties);
		}catch(AlreadyInitializedException e){
			Frontend.getInstance().displayInfo("Attempted to initialize already initialized config reader.");
		}catch(IOException e){
			Frontend.getInstance().displayWarning("Unable to load properties file. Error message: "+e.getMessage());
		}finally {
			connector = new Connector(new File(xmlConfigFile));
		}
		
		
	}
	
	/**
	 * Perform a surface scan of the given malware sample file using the registered anti-virus engines.
	 * 
	 * @param malwareSample path to the malware sample to be scanned.
	 * @return list of the scan results reported from the antivirus engines.
	 * @throws NoScannersLoadedException no scanners were loaded during initialization of the system, thus no engines are available to perform scan.
	 */
	public ArrayList<ScanResult> scan(String malwareSample){
		Frontend.getInstance().displayInfo("Performing scan on "+malwareSample+" using "+connector.getScanners().size()+" scan engines.");
		ArrayList<Scanner> scanners = connector.getScanners();
		
		ThreadHandler th = new ThreadHandler(scanners);
		ArrayList<ScanResult> results = th.performScan(new File(malwareSample));
		
		return results;
	}
	
	/**
	 * Attempt to update all scanner registered in the system.
	 * 
	 * @return list of update results.
	 */
	public ArrayList <AVUpdateResult> update() {
		// Store time for update in registry.
		Preferences.userRoot().putLong("LAST_UPDATE", System.currentTimeMillis());
		
		ArrayList<Scanner> scanners = connector.getScanners();

		
		ThreadHandler th = new ThreadHandler(scanners);
		ArrayList<AVUpdateResult> results = th.updateScanners();
		
		return results;
	}
	
	/**
	 * Get the date of the last call to PowerScan.update() from this client.
	 * 
	 * @return String representing the date of the last update() call.
	 * @see PowerScan#update()
	 */
	public static String getLastUpdate() throws NoMatchFoundException{
		long lastUpdateLong = Preferences.userRoot().getLong("LAST_UPDATE", -1);

		if(lastUpdateLong == -1){
			throw new NoMatchFoundException();
		}
		
		return new SimpleDateFormat("HH:mm dd/MM/yyyy").format(new Date(lastUpdateLong));		
	}
	
	/**
	 * Execute the supplied malware sample on all registered scanners. 
	 * 
	 * If a scanner does not support real time scanning, the Scanner object associated with it will not attempt to execute the malware.
	 * 
	 * @param malwareSample absolute or relative path to the sample to be executed.
	 * @return list of scan results resulting from the operation.
	 */
	public ArrayList<ScanResult> executeMalwareSample(String malwareSample){
		Frontend.getInstance().displayInfo("Executing "+malwareSample+" on "+connector.getScanners().size()+" virtual machines.");
		
		ThreadHandler th = new ThreadHandler(connector.getScanners());
		ArrayList<ScanResult> results = th.executeMalware(new File(malwareSample));
		
		return results;
	}
	
	
	/**
	 * Run the supplied malware sample in all the registered dynamic analysis tools
	 * 
	 * @param malwareSample path to the malware sample to be analyzed
	 * @return list of analysis tool result objects
	 */
	public ArrayList<DynamicAnalysisResult> executeInAnalysisTool(String malwareSample){
		Frontend.getInstance().displayInfo("Executing dynamic analysis on file " + malwareSample);
		
		ArrayList<Executor> executors = connector.getExecutors();
		ArrayList<DynamicAnalysisResult> analysisResults = new ArrayList<DynamicAnalysisResult>();
		
		// Sequential execution of tools on their respective VM
		for(Executor executor : executors) {
			try {
				ArrayList<DynamicAnalysisResult> result = executor.executeTools(new File(malwareSample), false);
				analysisResults.addAll(result);
			}
			catch(InvalidFileException ife) {
				Frontend.getInstance().displayError("Unable to read malwaresample on local system");
			}
			catch(CreateSnapshotException cse){
				Frontend.getInstance().displayWarning("Unable to create snapshot");
			}
			catch(FileCopyException fce) {
				Frontend.getInstance().displayError("Error while copying sample file to remote system");
			}
			catch(FileExecutionException fee) {
				Frontend.getInstance().displayError("Unable to execute analysis tool on remote system");
			}
		}
		
		// Revert all VMs to snapshot
		for(Executor executor : executors) {
			executor.revertToSnapshot();
		}
		
		return analysisResults;
		
	}
	
	/**
	 * Create a logger that logs to the given file for a class with the given name.
	 * 
	 * @param filename name of the file to log to. Can be both relative and absolute.
	 * @param className name of the class to create the Logger object for (should be full package path).
	 * @return instance of the logger with the requested parameters.
	 */
	public static Logger getLogger(String filename, String className) {
		// Get the Logger object
		Logger logger = Logger.getLogger(className);
		logger.setLevel(Level.ALL);
		
		// Direct logging to file.
		try{
			// If log directories doesn't exist, create them
	    	String dir = new File(filename).getParent();
	    	if(dir!=null)
	    		new File(dir).mkdirs();
	    	
	    	//Create individual files or overwrite?		
			if(!PropertiesReader.getInstance().getBool("log.overwrite", true)){
				// Create today's date in wanted format.
				String date = new SimpleDateFormat("HHmmssddMMyyyy").format(new Date());
				
				if(filename.substring(filename.lastIndexOf(System.getProperty("file.separator"))).contains(".")){
					filename = filename.substring(0, filename.lastIndexOf(".")) +"_"+ date + filename.substring(filename.lastIndexOf("."));
				}else{
					filename += date;
				}
				 
			}
	    	
			FileHandler fh = new FileHandler(filename);
			fh.setFormatter(new Formatter() {
				public String format( LogRecord record){
					return record.getLevel()+ " " +new Date(record.getMillis()).toString() + ": " +record.getMessage()+ "\n";
				}
			});
			logger.addHandler(fh);
			logger.setUseParentHandlers(false);
		}catch(IOException e) {
			Frontend.getInstance().displayError("Unable to write to log at "+filename+" for class "+className+" - redirecting to STD err.");
		}
		return logger;
	}
	
	public static void main(String[] args) throws Exception{

	}
	
}
