package edu.ntnu.item.jt.system;

import edu.ntnu.item.jt.VMware.*;
import edu.ntnu.item.jt.VMware.errors.*;
import edu.ntnu.item.jt.configloader.PropertiesReader;
import edu.ntnu.item.jt.system.errors.*;
import edu.ntnu.item.jt.system.configbeans.AVUpdateInfoBean;
import edu.ntnu.item.jt.system.configbeans.AVRealTimeScanBean;
import java.io.*;
import java.util.ArrayList;
import java.util.logging.Logger;

/**
 * Used for managing scan logic.
 * 
 * Uses a Guest OS and relevant info to perform scan logic and parse and return result.
 * 
 * @author Jran Vagnby Lillesand
 *
 */
public class Scanner implements Runnable{
	
	private GuestOS scanMachine;
	private String avEnginePath, parameters;
	private String remoteLogPath;
	private String name;
	private AVUpdateInfoBean avUpdateBean;
	private AVRealTimeScanBean realTimeScanBean;
	
	private LogFilter avScanFilter;	
	private Thread thread;
	
	// Arguments to be read by thread
	private File malwareSample;
	private boolean takeSnapshotBefore;
	private ScanThreadCallback scanThreadCallback;
	private AVUpdateThreadCallback avUpdateThreadCallback;
	private int operation;
	
	private static Logger log = PowerScan.getLogger(PropertiesReader.getInstance().getString("log.system.scanner", "logs"+System.getProperty("file.separator")+"scanner.log"), "edu.ntnu.item.jt.system.scanner");
	
	/**
	 * Indicates that the operation to be performed by the Scanner thread is a scan operation.
	 */
	public static final int OPERATION_SCAN = 1;
	
	/**
	 * Indicates that the operation to be performed by the Scanner thread is a malware execution operation.
	 */
	public static final int OPERATION_EXECUTE = 2;
	
	/**
	 * Indicates that the operation to be performed by the Scanner thread is to set the virtual machine in ready state.
	 */
	public static final int OPERATION_UPDATE = 3;
	
	/**
	 * Default value for the Scanner thread execution (no operation).
	 */
	public static final int OPERATION_NONE = 0;

		
	
	/**
	 * Overrides the Runnable run operation. 
	 * 
	 * Note that this runnable object should not be invoked directly, but using the appropriate functions
	 * (i.e. startThreadedScan()).
	 */
	public void run(){
		switch(operation){
		case OPERATION_SCAN:
			Frontend.getInstance().displayInfo(name+" starting scan.");
			ScanResult sr = new ScanResult(this);
			
			log(LOG_INFO, "Performing scan in thread.");
			try{
				//Perform scan without reverting to snapshot afterwards (the thread handles this itself)
				sr = scan(this.malwareSample, this.takeSnapshotBefore, false);
				log(LOG_INFO, "Scan result obtained.");


			}catch(InvalidFileException e){
				sr.setResult("Scan failed.");
				log(LOG_WARNING, "An exception occured during scan. Exception: "+e.getClass()+", error message: "+e.getMessage());
				Frontend.getInstance().displayWarning("Unable to read malware sample from local disk for "+name+". Error: "+e.getMessage());
			}catch(FileCopyException e){
				sr.setResult("Scan failed.");
				log(LOG_WARNING, "An exception occured during scan. Exception: "+e.getClass()+", error message: "+e.getMessage());
				Frontend.getInstance().displayWarning("Copy of malware sample to or result log file from guest OS for "+name+". A common reason for this error is that the scan operation failed to write the log output to file (i.e. the file was not created). Error: "+e.getMessage());
			}catch(CreateSnapshotException e){
				sr.setResult("Scan failed.");
				log(LOG_WARNING, "An exception occured during scan. Exception: "+e.getClass()+", error message: "+e.getMessage());
				Frontend.getInstance().displayWarning("Failed to take snapshot prior to performing scan for "+name+". Error: "+e.getMessage());
			}catch(FileExecutionException e){
				sr.setResult("Scan failed.");
				log(LOG_WARNING, "An exception occured during scan. Exception: "+e.getClass()+", error message: "+e.getMessage());
				Frontend.getInstance().displayWarning("Unable to execute anti-virus engine executable for "+name+". Error: "+e.getMessage());
			}catch(IOException e){
				sr.setResult("Scan failed.");
				log(LOG_WARNING, "An exception occured during scan. Exception: "+e.getClass()+", error message: "+e.getMessage());
				Frontend.getInstance().displayWarning("Writing of log file to temprorary directory on local machine failed for "+name+". Error: "+e.getMessage());
			}catch(Exception e){
				sr.setResult("Scan failed.");
				log(LOG_WARNING, "An exception occured during scan. Exception: "+e.getClass()+", error message: "+e.getMessage());
				Frontend.getInstance().displayWarning("Unexpected exception "+e.getClass()+" occured for "+name+". Error message: "+e.getMessage());
			}finally{
				
				scanThreadCallback.reportResult(sr);
				
				try{
					log(LOG_INFO, "Reverting to snapshot.");
					this.revertToSnapshot(true);
					Frontend.getInstance().displayInfo("Successfully revert virtual machine to snapshot for "+name+".");
					log(LOG_INFO, "Successfully reverted to snapshot.");
				}catch(Exception f){
					Frontend.getInstance().displayError("Failed to revert to snapshot on virtual machine running "+name+". Please ensure that the VM is in ready state before executing new scan.");
					log(LOG_SEVERE, "Failed to revert to snapshot. State may be invalid.");
				}
				
			}
			
			// reset used variables:
			this.scanThreadCallback=null;
			operation=Scanner.OPERATION_NONE;
			this.malwareSample=null;
			break;
			
		case OPERATION_EXECUTE:
			Frontend.getInstance().displayInfo(name+" starting execution of malware sample at "+malwareSample.getAbsolutePath());
			log(LOG_INFO, "Starting execution in thread.");
			ScanResult result = new ScanResult(this);
			
			// This boolean is used to prevent revertToSnapshot from being called 
			boolean executed = false;
			
			try{ // This try catches failures that may occur during exceution.
				
				try{ // This try catches OperationNotSupportedException.
					
					// Execute malware sample without reverting to snapshot afterwards (the thread handles this itself).
					result = this.executeMalwareSample(malwareSample, takeSnapshotBefore, false);
					executed = true;
					log(LOG_INFO, "Scan result obtained.");
					
				}catch(OperationNotSupportedException e){
					// Operation not supported, report back generic result
					result.setScanTime(0);
					result.setResult("Real-time scanning not supported.");
					executed = false; // Already done implicitly, but to done again to prevent errors if code is modified.
				}
				
			}catch(IOException e){
				result.setResult("Malware execution failed - IO Exception occured.");
				log(LOG_WARNING, "An exception occured while trying to execute malware on virtual machine. Exception: "+e.getClass()+", error message: "+e.getMessage());
				Frontend.getInstance().displayWarning("Writing of log file to temprorary directory on local machine failed for virtual machine running "+name+". Error: "+e.getMessage());
			}catch(FileCopyException e){
				result.setResult("Malware execution failed.");
				log(LOG_WARNING, "An exception occured while trying to execute malware on virtual machine. Exception: "+e.getClass()+", error message: "+e.getMessage());
				Frontend.getInstance().displayWarning("Copy of malware sample to or result log file from guest OS for virtual machine running "+name+". A common reason for this error is that the scan operation failed to write the log output to file (i.e. the file was not created). Error: "+e.getMessage());
			}catch(InvalidFileException e){
				result.setResult("Malware execution failed.");
				log(LOG_WARNING, "An exception occured while trying to execute malware on virtual machine. Exception: "+e.getClass()+", error message: "+e.getMessage());
				Frontend.getInstance().displayWarning("Unable to read malware sample from local disk virtual machine running "+name+". Error: "+e.getMessage());
			}catch(CreateSnapshotException e){
				result.setResult("Malware execution failed.");
				log(LOG_WARNING, "An exception occured while trying to execute malware on virtual machine. Exception: "+e.getClass()+", error message: "+e.getMessage());
				Frontend.getInstance().displayWarning("Failed to take snapshot prior executing malware on machine running "+name+". Error: "+e.getMessage());
			}catch(Exception e){
				result.setResult("Malware execution failed.");
				log(LOG_WARNING, "An exception occured while trying to execute malware on virtual machine. Exception: "+e.getClass()+", error message: "+e.getMessage());
				Frontend.getInstance().displayWarning("Unexpected exception "+e.getClass()+" occured virtual machine running "+name+". Error message: "+e.getMessage());
			}finally{
				
				scanThreadCallback.reportResult(result);
				
				if(executed){
					try{
						log(LOG_INFO, "Reverting to snapshot.");
						this.revertToSnapshot(true);
						log(LOG_INFO, "Successfully reverted to snapshot.");
						Frontend.getInstance().displayInfo("Reverted virtual machine running "+name+" to snapshot.");
					}catch(Exception f){
						Frontend.getInstance().displayError("Failed to revert to snapshot on virtual machine running "+name+". THIS MAY INDICATE THAT THE MALWARE IS STILL RUNNING ON THE VIRTUAL MACHINE.");
						log(LOG_SEVERE, "Failed to revert to snapshot. State may be invalid.");
					}
				}
				
			}
			
			// Reset used variables.
			this.scanThreadCallback=null;
			operation=Scanner.OPERATION_NONE;
			this.malwareSample=null;
			break;
			
		case OPERATION_UPDATE:
			Frontend.getInstance().displayInfo(name+" starting update operation.");

			log(LOG_INFO, "Attempting to update AV engine.");
			
			avUpdateThreadCallback.reportResult(updateAV());
			
			// Reset variables used in thread
			avUpdateThreadCallback = null;
			operation = Scanner.OPERATION_NONE;
			break;
			
			// TODO: Insert any other operations to be executed in parallel here.
		default:
			// do nothing, but report error.
			Frontend.getInstance().displayWarning("Attempted to run thread with engine "+this.name+", but no operation was specified.");
			log(LOG_WARNING, "Attempted to run thread with no operation specified.");	
			break;
		}
		
		// remove reference to thread after use (allow any resources to be freed up).
		thread=null;
	}
	
	/**
	 * Constructor for scanner class.
	 * 
	 * @param scanMachine the virtual machine to be used for scanning.
	 * @param name the name of the running antivirus engine.
	 * @param avEnginePath path to the antivirus executable on the remote machine.
	 * @param parameters required parameters to the antivirus executable on the remote machine (including location of database files, output directory for log files etc). Only relevant for explicit scans. 
	 * @param remoteLogPath path to log files resulting from the scan operation.
	 * @param avScanFilter filter to be used for identifying the result from the log file of a scan operation.
	 * @param avUpdateBean bean containing the information related to performing automatic update of the virus definitions for this scanner. If automatic update is not supported, supply null.
	 * @param realTimeScanBean containing information needed to utilize the real time scanning capability of this scanner. If real time scanning is not supported, supply null.
	 */
	public Scanner(GuestOS scanMachine, String name, String avEnginePath, String parameters, String remoteLogPath, LogFilter avScanFilter, AVUpdateInfoBean avUpdateBean, AVRealTimeScanBean realTimeScanBean){
		this.scanMachine = scanMachine;
		this.name = name;
		this.avEnginePath = avEnginePath;
		this.parameters = parameters;
		this.remoteLogPath = remoteLogPath;
		this.avScanFilter = avScanFilter;
		this.avUpdateBean = avUpdateBean;
		this.realTimeScanBean = realTimeScanBean;
		log(LOG_INFO, "Instantiated scanner object.");
	}
	
	public String toString(){
		return name+" running on "+ scanMachine.getHost();
	}
	

	/**
	 * Launch a separate thread that performs scan and reports back to the given callback object.
	 * 
	 * More specifically, the function sets some instance variables which starts a thread which 
	 * performs scan operations based on these variables. 
	 * 
	 * @param localSamplePath specifies the malware sample to be scanned.
	 * @param takeSnapshotBeforeScan specifies whether to take a snapshot before the scan is performed (resource demanding).
	 * @param scanThreadCallback specifies which object the thread should report back to upon finishing the scan.
	 */
	public void startThreadedScan(File localSamplePath, boolean takeSnapshotBeforeScan, ScanThreadCallback scanThreadCallback){
		//Set necessary parameters before scan.
		operation = Scanner.OPERATION_SCAN;
		this.malwareSample = localSamplePath;
		this.takeSnapshotBefore = takeSnapshotBeforeScan;
		this.scanThreadCallback = scanThreadCallback;
		
		// Object configured, start scan.
		thread = new Thread(this);
		thread.start();
	}
	
	
	/**
	 * Attempt to update the AV engine for this scanner.
	 * 
	 * @return result description of the operation.
	 */
	public AVUpdateResult updateAV(){
		log(Scanner.LOG_INFO, "Starting update of AV definitions.");
		AVUpdateResult result = new AVUpdateResult(this);
		try{
			log(Scanner.LOG_INFO, "Powering up and logging in to virtual machine.");
			// Powerup and login to scan machine. This should already be done by the connector, but do it anyways to make sure it's done.
			scanMachine.powerOn(true, PropertiesReader.getInstance().getInt("vmware.tools.timeout", 60)); //always wait for tools (required for later operations).
			scanMachine.login();
			// Unless exception was thrown, these operations can be assumed to have been successful.

			if(avUpdateBean != null){
				try{
					
					Frontend.getInstance().displayInfo("Starting update of virus definitions for "+name+".");
					
					//System.out.println(updateExecutable + " " + updateParameters);
					log(Scanner.LOG_INFO, "Executing update executable \""+avUpdateBean.getUpdateExecutable()+"\" with parameters \""+ avUpdateBean.getUpdateParameters()+"\"");
					scanMachine.runProgramInGuest(avUpdateBean.getUpdateExecutable(), avUpdateBean.getUpdateParameters());
					
					// If output file is produced, copy it to local disk and attempt to filter it.
					ArrayList<String> scanLog = null;
					try{
						if(avUpdateBean.getUpdateLogPath() != null) {
							log(Scanner.LOG_INFO, "Creating temp file for temporary storage of update result log.");
							File temp = File.createTempFile("updateresult", "res");
							temp.deleteOnExit();
							log(Scanner.LOG_INFO, "File created at "+temp.getAbsolutePath());
							
							log(Scanner.LOG_INFO, "Copying file from remote path "+avUpdateBean.getUpdateLogPath()+" to local temp file.");
							scanMachine.copyFileFromGuest(avUpdateBean.getUpdateLogPath(), temp.getAbsolutePath());
							
							//write result log to Scanner log
							writeUpdateResultToLog(temp);
							
							//Try to delete log file from guest OS to avoid when parsing log after next update.
							try{
								log(Scanner.LOG_INFO, "Deleting log file from guest OS at path "+avUpdateBean.getUpdateLogPath()+".");
								scanMachine.runProgramInGuest("c:\\windows\\system32\\cmd.exe", "/C del "+avUpdateBean.getUpdateLogPath());
							}catch(FileExecutionException e){
								log(Scanner.LOG_WARNING, "Unable to delete log file, error: "+e.getMessage());
								Frontend.getInstance().displayWarning("Unable to delete scan log file for "+name+". Please review error and delete manually to avoid erroneous reporting of later scans. Error: "+e.getMessage());
							}
							
							//Filter file
							if(avUpdateBean.getUpdateSuccessIndicator() != null){
								try{
									log(Scanner.LOG_INFO, "Filtering update result log.");
									scanLog = new LogFilter(avUpdateBean.getUpdateSuccessIndicator()).filter(temp);
									//If this point is reached, a match of the success indicator was found => was successful
									result.setResultState(AVUpdateResult.UPDATE_SUCCESSFUL);
									result.setResultString(scanLog);
								}catch(NoMatchFoundException e){
									log(Scanner.LOG_WARNING, "No match found in result log, returning full update result log to user for investigation.");
									
									//If no match is found, return the full scan log for further investigation
									try{
										result.setResultState(AVUpdateResult.SUCCESS_INDICATOR_NOT_FOUND);
										result.setResultString(new LogFilter().filter(temp));
									}catch(NoMatchFoundException f){
										log(Scanner.LOG_WARNING, "Update result log file is empty.");
										// Update log file is empty.
										
										result.setResultState(AVUpdateResult.NO_LOG_FOUND);
									}
								}
							}else{
								//Return whole log file.
								try{
									log(Scanner.LOG_INFO, "Scan engine does not support automatic update result log analysis, displaying whole log to user.");
									scanLog = new LogFilter().filter(temp);
									result.setResultState(AVUpdateResult.FULL_LOG_RETURNED);
									result.setResultString(scanLog);
								}catch(NoMatchFoundException e){
									// No log found
									log(Scanner.LOG_WARNING, "Update result log file is empty.");
									result.setResultState(AVUpdateResult.NO_LOG_FOUND);
								}
							}
						}else{
							//Update operation does not support logging
							log(Scanner.LOG_INFO, "Scan engine does not support logging of update operation.");
							result.setResultState(AVUpdateResult.LOG_NOT_SUPPORTED);
						}
						
					}catch(IOException e){
						log(Scanner.LOG_WARNING, "Unable to verify update result as result log could not be written to local file. Error: "+e.getMessage());
						result.setResultState(AVUpdateResult.NO_LOG_FOUND);
						Frontend.getInstance().displayError("Unable to verify update result for virtual machine "+name+" as result log could not be written to local file.");
					}catch(FileCopyException e){
						log(Scanner.LOG_WARNING, "Unable to verify update result as result log could not be copied to local system. Error: "+e.getMessage());
						result.setResultState(AVUpdateResult.NO_LOG_FOUND);
						Frontend.getInstance().displayError("Unable to verify update result for virtual machine "+name+" as result log could not be copied to local system.");
					}
					
					// As long as update file was executed, the operation MAY have been successful. Take snapshot.
					try{
						log(Scanner.LOG_INFO, "Taking snapshot.");
						scanMachine.takeSnapshot("PowerScan snapshot", "Automatically taken snapshot by PowerScan after (presumably successful) update of virus definitions.", false);
						
						Frontend.getInstance().displayInfo("Automatic snapshot taken for VM running "+name+".");
					}catch(CreateSnapshotException e){
						log(Scanner.LOG_SEVERE, "Snapshot creation failed. Error: "+e.getMessage());
						Frontend.getInstance().displayError("Unable to take snapshot in virtal machine running "+name+" after performing AV update. Please review the virtual machine and make sure that it is a valid state (even after reverting to snapshot).");
					}
				}catch(FileExecutionException e){
					log(Scanner.LOG_SEVERE, "Unable to execute file "+avUpdateBean.getUpdateExecutable()+". Error: "+e.getMessage());
					Frontend.getInstance().displayWarning("Update of "+name+" failed. Could not execute "+avUpdateBean.getUpdateExecutable()+": "+e.getMessage());
				}
			}else{
				log(Scanner.LOG_INFO, "Scan engine does not support automatic updates");
				Frontend.getInstance().displayInfo("Automatic update not supported for "+name+".");
				result.setResultState(AVUpdateResult.UPDATE_NOT_SUPPORTED);
			}
			
			//if not successful, report this to the user and give needed info to perform operation manually.
			
		}catch(PowerUpFailureException e){
			log(Scanner.LOG_SEVERE, "Unable to power up virtual machine.");
			Frontend.getInstance().displayError("Unable to power up virtual machine for "+name+". Please check the state of the virtual machine manually.");
		}catch(LoginFailureException e){
			log(Scanner.LOG_SEVERE, "Unable to log in to guest OS on Virtual Machine.");
			Frontend.getInstance().displayError("Unable to log in to virtual machine running "+name+". Please check the state of the virtual machine manually.");
		}
		
		return result;
		
	}
	
	/**
	 * Start a AV engine update in a thread.
	 * 
	 * Sets the needed parameters, and starts a thread for this object.
	 * 
	 * @param callback
	 * @see Scanner#updateAV()
	 */
	public void startThreadedAVUpdate(AVUpdateThreadCallback callback){
		
		//Set necessary parameters for thread.
		operation = Scanner.OPERATION_UPDATE;
		avUpdateThreadCallback = callback;
		
		thread = new Thread(this);
		thread.start();
	}
	
	
	/**
	 * Performs a scan operation on a remote machine.
	 * 
	 * This involves the following operations:
	 * <ol>
	 * 	<li>Optionally take snapshot of current virtual machine state</li>
	 *  <li>Copy file from local machine to guest OS (virtual machine)</li>
	 * 	<li>Run antivirus engine with given parameters in guest OS with the copied file as input.</li>
	 * 	<li>Copy resulting scan operation log file from Guest OS to local machine </li>
	 * 	<li>Optionally reverting to snapshot after scan (recommended)</li>
	 * 	<li>Reading and filtering log</li>
	 * 	<li>Analyzing result</li>
	 * <ul>
	 * 
	 * Note that this operation presumes that the Guest OS in use is powered on and logged in with a user of sufficient privilege.
	 * 
	 * Also note that occurences of the string "$samplePath" in the parameters are replaced with the sample path on the remote machine (guest OS).
	 * 
	 * @param localSamplePath full path to the local sample. Note that if local machine uses backslash as denominator, these must be escaped (i.e. <code> c:\sample.exe </code> must be supplied as <code> c:\\sample.exe </code>).
	 * @param takeSnapshotBeforeScan specifies whether to take a snapshot of the current machine state before performing scan. 
	 * @param revertToSnapshotAfterScan specifies whether to revert to snapshot after performing scan (recommended).
	 * @return object containing aggregated scan result information. 
	 * @throws InvalidFileException if the local file sample can't be read.
	 * @throws FileCopyException if file copying failed (either of sample to guest os or scan log file from guest os).
	 * @throws CreateSnapshotException if snapshot creation failed.
	 * @throws FileExecutionException if file (antivirus engine executable) execution failed.
	 * @throws IOException if writing of log file to temp directory on local machine failed.
	 * @see LogFilter
	 * @see ScanResult
	 * 
	 */
	public ScanResult scan(File localSamplePath, boolean takeSnapshotBeforeScan, boolean revertToSnapshotAfterScan) throws InvalidFileException, FileCopyException, CreateSnapshotException, FileExecutionException, IOException, DirectoryCreationException{
		
		log(LOG_INFO, "Starting scan of file "+localSamplePath.getAbsolutePath()+ " take snapshot before scan: "+takeSnapshotBeforeScan+". Restore snapshot after scan: "+revertToSnapshotAfterScan);
		long startTime = System.currentTimeMillis();
		
		// Check that files are okay
		if(!localSamplePath.isFile() || !localSamplePath.canRead()){
			log(LOG_WARNING, "Unable to read file. Aborting scan.");
			throw new InvalidFileException("Unable to read malware sample file.");
		}
		
		//Find filename of file to be scanned
		String fileName = localSamplePath.getAbsolutePath().substring(localSamplePath.getAbsolutePath().lastIndexOf(File.separatorChar)+1);
		if(VMwareServer.debug) System.out.println("Attempting to scan file sample "+fileName+".");
		
		// Take snapshot?
		if(takeSnapshotBeforeScan){
			log(LOG_INFO, "Taking snapshot of virtual machine state.");
			scanMachine.takeSnapshot("Before Scan", "Snapshot automatically taken prior to performing scan.", false);
			log(LOG_INFO, "Snapshot successfully taken.");
		}
					
		// Read location to which to copy the file.
		String remotePath = PropertiesReader.getInstance().getString("scanner.executionpath", "c:\\");
		
		// Attempt to create directory in case it does not exist.
		// This is not supported in the current versions of VIX.
		// scanMachine.createDirectory(remotePath);
		
		log(LOG_INFO, "Copying local file to "+remotePath+fileName+" on guest OS.");
		//long copyTimeStart = System.currentTimeMillis();
		scanMachine.copyFileToGuest(localSamplePath.getAbsolutePath(), remotePath+fileName);
		//if(Frontend.debug) System.out.println("Time to copy file: "+ (System.currentTimeMillis()- copyTimeStart) / 1000);
		
		try{
			// Perform scan.
			if(VMwareServer.debug) System.out.println("Running: " +avEnginePath+ " " +parameters +" "+remotePath+fileName);
			
			// Replace any instance of $samplePath in the parameters with actual sample path on the remote system
			parameters = parameters.replace("$samplePath", remotePath+fileName);
			
			CallbackProc callback = new CallbackProc(scanMachine);
			log(LOG_INFO, "Performing scan: "+avEnginePath+" "+ parameters);
			callback.setJobHandle(scanMachine.runProgramInGuest(avEnginePath, parameters, callback));
			
			int timeout = PropertiesReader.getInstance().getInt("scanner.timeout", 25);
			double pollingInterval = PropertiesReader.getInstance().getDouble("polling.interval.minor", 0.250);
			
			log(LOG_INFO, "Polling for finish of scan. Polling interval: "+pollingInterval+". Timeout: "+timeout);
			
			// Poll for completion of file execution
			for(int i = 0; i < timeout/ pollingInterval; i++){
				//if(Frontend.debug) System.out.println("polling "+ i +"/"+ timeout/(double) (pollingInterval) +" "+name);
				
				if(callback.isFinished()){
					log(LOG_INFO, "Received callback - scan finished.");
					break;
				}
				
				try{
					Thread.sleep((long) (pollingInterval*1000));
				}catch(InterruptedException e){}
			}
			
			if(!callback.isFinished()){
				log(LOG_WARNING, "Did not receive callback before timeout. Scan timed out.");
				throw new TimedOutException("AV scan execution timed out.");
			}
			
			//Old code using blocking program execution (error prone).
			//scanMachine.runProgramInGuest(avEnginePath, parameters +" "+"c:\\"+fileName);
			
			
			// Get result.
			File result = File.createTempFile("scanresult", "res");
			result.deleteOnExit();
			log(LOG_INFO, "Created local file to store scan result: "+result.getAbsolutePath());
			
			// Copy remote log file to result for parsing
			log(LOG_INFO, "Copying scan result from remote path "+remoteLogPath+" to local path "+ result.getAbsolutePath());
			scanMachine.copyFileFromGuest(remoteLogPath, result.getAbsolutePath());
			
			// Try to write result to log
			writeScanResultToLog(result);
			
			try{
				if(revertToSnapshotAfterScan){
					log(LOG_INFO, "Reverting to snapshot in blocking operation.");
					//If snapshot has been taken this session, do not attempt to load first, as the wrong handle will be returned. The correct one is set during snapshot creation. Else, load snapshot.
					if(!takeSnapshotBeforeScan)
						scanMachine.loadSnapshot();
					
					scanMachine.revertToSnapshot(true);
				}
			}catch(NoSnapshotAvailableException e){
				if(VMwareServer.debug) System.out.println("Failed to revert to snapshot after scanning: No snapshot was available.");
				log(LOG_WARNING, "Failed to revert to snapshot after scanning: No snapshot was available.");
			}catch(NoSnapshotLoadedException e){
				if(VMwareServer.debug) System.out.println("Failed to revert to snapshot after scanning: No snapshot could be loaded available.");
				log(LOG_WARNING, "Failed to revert to snapshot after scanning: No snapshot could be loaded available.");
			}catch(RevertToSnapshotFailedException e){
				if(VMwareServer.debug) System.out.println("Failed to revert to snapshot after scanning: An error occured while reverting to snapshot.");
				log(LOG_WARNING, "Failed to revert to snapshot after scanning: An error occured while reverting to snapshot.");
			}
			
			// Build scanresult
			ScanResult scanResult = new ScanResult(this);
			long lapsedTime = System.currentTimeMillis() - startTime;
			scanResult.setScanTime(lapsedTime);
			
			try{
				ArrayList<String> results = avScanFilter.filter(result);
				scanResult.setResults(results);
				log(LOG_INFO, "Read results from scan log.");
			}catch(NoMatchFoundException e){
				scanResult.setResult("No infection detected.");
				log(LOG_INFO, "No infection read from scan log.");
			}
			
			return scanResult;
			
		}catch(TimedOutException e){
			log(LOG_INFO, "Generating scan result to report timeout.");
			ScanResult scanResult = new ScanResult(this);
			long lapsedTime = System.currentTimeMillis() - startTime;
			System.out.println(lapsedTime);
			scanResult.setScanTime(lapsedTime);
			scanResult.setResult("Operation timed out.");
			return scanResult;
		}
		
	}
	
	/**
	 * Creates and starts a thread which will execute a malware sample on the virtual machine associated with this Scanner object.
	 * 
	 * The Thread will will execute untill completion and report back to the provided callback.
	 * 
	 * @param localSamplePath reference to the malware sample.
	 * @param takeSnapshotBeforeScan boolean specifying whether a snapshot should be taken before executing the malware. This is only recommended if no clean snapshot is present on the virtual machine.
	 * @param callback object to be called when execution is finished (or timed out).
	 * @see Scanner#scan(File, boolean, boolean)
	 */
	public void startThreadedMalwareExecution(File localSamplePath, boolean takeSnapshotBeforeScan, ScanThreadCallback callback){
		//Set necessary parameters before scan.
		operation = Scanner.OPERATION_EXECUTE;
		this.malwareSample = localSamplePath;
		this.takeSnapshotBefore = takeSnapshotBeforeScan;
		this.scanThreadCallback = callback;
		
		// Object configured, start scan.
		thread = new Thread(this);
		thread.start();
	}
	
	/**
	 * Executes the supplied malware sample on the virtual machine associated with this scanner.
	 * 
	 * The malware execution process constists of the following phases:
	 * <ul>
	 * 	<li> Optionally take snapshot of system before doing any processing. </li>
	 * 	<li> Copy malware sample to some location on virtual machine. </li>
	 *  <li> Execute the malware sample.</li>
	 *  <li> Wait for the first event to occur of either a timeout or malware execution completion. </li>
	 *  <li> Copy real-time anti-virus scanner log file to local drive. </li>
	 *  <li> Optionally revert virtual machine to previously taken snapshot (if this is not done here, it should be done manually by some other means). </li>
	 *  <li> Analyze the log file and extract results. </li>
	 * </ul>
	 * 
	 * @param malwareSample malware sample to be executed.
	 * @param takeSnapshotBeforeExecution boolean stating whether snapshot should be taken prior to copying and execution.
	 * @param revertToSnapshotAfterExecution boolean stating whether snapshot revertion should be performed after executing malware.
	 * @return result of the real time scan operation.
	 * @throws IOException if an error occured while processing the local copy of the log file created at the client (the local machine).
	 * @throws InvalidFileException if the malware sample could not be read.
	 * @throws CreateSnapshotException if the snapshot could not be created.
	 * @throws FileCopyException if an exception occured while copying the malware sample to the 
	 */
	public ScanResult executeMalwareSample(File malwareSample, boolean takeSnapshotBeforeExecution, boolean revertToSnapshotAfterExecution) throws IOException, InvalidFileException, CreateSnapshotException, FileCopyException, OperationNotSupportedException{

		log(LOG_INFO, "Executing malware sample "+malwareSample.getAbsolutePath()+ " take snapshot before execution: "+takeSnapshotBeforeExecution+". Restore snapshot after scan: "+revertToSnapshotAfterExecution);
		
		///Check if supported.
		if(realTimeScanBean == null){
			log(LOG_INFO, "Malware execution aborted, real time scanning not supported.");
			throw new OperationNotSupportedException();
		}		
		
		long startTime = System.currentTimeMillis();
		
		// Check that files are okay
		if(!malwareSample.isFile() || !malwareSample.canRead()){
			log(LOG_WARNING, "Unable to read malware sample. Aborting.");
			throw new InvalidFileException("Unable to read malware sample file.");
		}
		
		//Find filename of malware file
		String fileName = malwareSample.getAbsolutePath().substring(malwareSample.getAbsolutePath().lastIndexOf(File.separatorChar)+1);
		if(VMwareServer.debug) System.out.println("Attempting to execute file sample "+fileName+".");
		
		// Take snapshot?
		if(takeSnapshotBeforeExecution){
			log(LOG_INFO, "Taking snapshot of virtual machine state.");
			scanMachine.takeSnapshot("Before malware execution", "Snapshot automatically taken prior executing malware.", false);
			log(LOG_INFO, "Snapshot successfully taken.");
		}
					
		// Read location to which to copy the file.
		String remotePath = PropertiesReader.getInstance().getString("scanner.executionpath", "c:\\");
		
		// Attempt to create directory in case it does not exist.
		// TODO: This is not supported in the current versions of VIX. IT IS HOWEVER DOABLE IN A PLATFORM SPECIFIC MANNER: SIMILARILY SO FOR SCAN()!! (use mkdir in CMD)
		// scanMachine.createDirectory(remotePath);
		
		log(LOG_INFO, "Copying local malware sample to "+remotePath+fileName+" on guest OS.");
		long copyTimeStart = System.currentTimeMillis();
		scanMachine.copyFileToGuest(malwareSample.getAbsolutePath(), remotePath+fileName);
		if(Frontend.debug) System.err.println("Time to copy file: "+ (System.currentTimeMillis()- copyTimeStart) / 1000);
		
		// Execute sample.
		CallbackProc callback = new CallbackProc(scanMachine);
		log(LOG_INFO, "Executing malware sample: "+remotePath+fileName);
		callback.setJobHandle(scanMachine.runProgramInGuest(remotePath+fileName, "", callback));
		
		int timeout = PropertiesReader.getInstance().getInt("malware.execution.timeout", 25);
		double pollingInterval = PropertiesReader.getInstance().getDouble("polling.interval.minor", 0.250);
		
		log(LOG_INFO, "Polling for finish of malware execution. Polling interval: "+pollingInterval+". Timeout: "+timeout);
		
		// Poll for completion of file execution
		for(int i = 0; i < timeout/ pollingInterval; i++){
			//if(Frontend.debug) System.out.println("polling "+ i +"/"+ timeout/(double) (pollingInterval) +" "+name);

			if(callback.isFinished()){
				log(LOG_INFO, "Received callback - malware sample process terminated or was interrupted by real-time AV software.");
				break;
			}
			
			try{
				Thread.sleep((long) (pollingInterval*1000));
			}catch(InterruptedException e){}
		}
		
		// Get result.
		File result = File.createTempFile("scanresult", "res");
		result.deleteOnExit();
		log(LOG_INFO, "Created local file to store AV log: "+result.getAbsolutePath());
		
		// Copy remote log file to result for parsing
		log(LOG_INFO, "Copying scan result from remote path "+realTimeScanBean.getLogFilePath()+" to local path "+ result.getAbsolutePath());
		scanMachine.copyFileFromGuest(realTimeScanBean.getLogFilePath(), result.getAbsolutePath());
		
		// Try to write result to log
		writeScanResultToLog(result);
		
		if(revertToSnapshotAfterExecution){
			try{
				log(LOG_INFO, "Reverting to snapshot in blocking operation.");
				//If snapshot has been taken this session, do not attempt to load first, as the wrong handle will be returned. The correct one is set during snapshot creation. Else, load snapshot.
				if(!takeSnapshotBeforeExecution)
					scanMachine.loadSnapshot();
				
				scanMachine.revertToSnapshot(true);
				// In practice, this block is never used... Same for scan().
			}catch(NoSnapshotAvailableException e){
				Frontend.getInstance().displayError("Failed to revert to snapshot after malware execution for scanner "+name+": No snapshot was available. NOTE THAT THE MALWARE MAY STILL BE ACTIVE ON THE SYSTEM.");
				log(LOG_WARNING, "Failed to revert to snapshot after scanning: No snapshot was available.");
			}catch(NoSnapshotLoadedException e){
				Frontend.getInstance().displayError("Failed to revert to snapshot after scanning for scanner "+name+": No snapshot could be loaded available. NOTE THAT THE MALWARE MAY STILL BE ACTIVE ON THE SYSTEM.");
				log(LOG_WARNING, "Failed to revert to snapshot after scanning: No snapshot could be loaded available.");
			}catch(RevertToSnapshotFailedException e){
				Frontend.getInstance().displayError("Failed to revert to snapshot after scanning for scanner "+name+": An error occured while reverting to snapshot. NOTE THAT THE MALWARE MAY STILL BE ACTIVE ON THE SYSTEM.");
				log(LOG_WARNING, "Failed to revert to snapshot after scanning: An error occured while reverting to snapshot.");
			}
			
		}
			
		// Create filter
		LogFilter filter = new LogFilter(realTimeScanBean.getResultIdentifier());
		
		// Build scanresult
		ScanResult scanResult = new ScanResult(this);
		long lapsedTime = System.currentTimeMillis() - startTime;
		scanResult.setScanTime(lapsedTime);
		
		try{
			ArrayList<String> results = filter.filter(result);
			scanResult.setResults(results);
			log(LOG_INFO, "Read results from scan log.");
		}catch(NoMatchFoundException e){
			scanResult.setResult("No infection detected.");
			log(LOG_INFO, "No infection read from scan log.");
		}
		
		return scanResult;	
	}
	
	/**
	 * Getter for the AVE name associated with this scanner object.
	 * 
	 * @return name of the AVE associated with this scanner object.
	 */
	public String getName() {
		return name;
	}
	
	/**
	 * Get the scan machine (guest OS) used by this scanner.
	 * 
	 * @return guest OS used by the scanner.
	 */
	public GuestOS getScanMachine(){
		return scanMachine;
	}
	
	
	/**
	 * Load a snapshot and revert to it.
	 * 
	 * @param blocking specifies whether the call should be made in a blocking manner.
	 */
	public void revertToSnapshot(boolean blocking) throws NoSnapshotLoadedException, RevertToSnapshotFailedException, NoSnapshotAvailableException{
		scanMachine.revertToSnapshot(blocking);
	}
	
	
	/**
	 * Writes a scan result file to log.
	 * 
	 * @param file scan result.
	 */
	private void writeScanResultToLog(File file){
		try{
			String logEntry="";
			BufferedReader br = new BufferedReader(new InputStreamReader(new FileInputStream(file)));
			while(br.ready()){
				logEntry += br.readLine()+"\n";
			}
			
			log(LOG_INFO, "Scan result log file: \n"+logEntry);
		}catch(Exception e){
			log(LOG_WARNING, "Unable to write result file to log: "+e.getMessage());
		}
	}
	
	/**
	 * Writes a scan result file to log.
	 * 
	 * @param file scan result.
	 */
	private void writeUpdateResultToLog(File file){
		try{
			String logEntry="";
			BufferedReader br = new BufferedReader(new InputStreamReader(new FileInputStream(file)));
			while(br.ready()){
				logEntry += br.readLine()+"\n";
			}
			
			log(LOG_INFO, "Scan result log file: \n"+logEntry);
		}catch(Exception e){
			log(LOG_WARNING, "Unable to write result file to log: "+e.getMessage());
		}
	}
	
	
	/**
	 * Private logger to assist in parsing log messages .
	 * 
	 * @param severity severity level of the occurence.
	 * @param message log message.
	 */
	private void log(int severity, String message){
		
		switch(severity){
		case LOG_INFO:
			log.info("Scanner '"+ name +"': " +message);
			break;
		case LOG_WARNING:
			log.warning("Scanner '"+ name +"': " +message);
			break;
		case LOG_CONFIG:
			log.config("Scanner '"+ name +"': " +message);
			break;
		case LOG_SEVERE:
			log.severe("Scanner '"+ name +"': " +message);
			break;
		case LOG_ENTERING:
			//TODO: Implement this?
			break;
		case LOG_EXITING:
			//TODO: Implement this?
			break;
		}
	}
	
	/*
	 * Log levels
	 */
	private static final int LOG_INFO = 0;
	private static final int LOG_WARNING = 1;
	private static final int LOG_CONFIG = 2;
	private static final int LOG_SEVERE = 3;
	private static final int LOG_ENTERING = 4;
	private static final int LOG_EXITING = 5;

	/**
	 * @param args
	 */
	public static void main(String[] args) throws Exception {
		
		String vmxPath = "/var/lib/vmware-server/Virtual Machines/winxp_0/Windows XP Professional.vmx";
        String hostName = "dhcp208-210.ed.ntnu.no";
        int portNumber = 902;
        
    	VMwareServer vws = new VMwareServer(hostName, portNumber);
        vws.connect("joeranvl", "feitfrank");
        
        GuestOS winxp1 = vws.openVM(vmxPath, "thomas", "thomas");
        winxp1.powerOn(true, 0);
        
        winxp1.login();
        
        //Scanner clamWin = new Scanner(winxp1, "C:\\Program Files\\ClamWin\\bin\\clamscan.exe", "--database=\"C:\\Documents and Settings\\All Users\\.clamwin\\db\\main.inc\" --log=c:\\result.log", "c:\\result.log");
        //File file = clamWin.scan(new File("c:\\eicar.com"), true, true);
        
        //System.out.println(file);
	}

}
