package edu.ntnu.item.jt.system;

import java.util.ArrayList;
import java.io.File;

import edu.ntnu.item.jt.VMware.errors.RevertToSnapshotFailedException;
import edu.ntnu.item.jt.configloader.PropertiesReader;

/**
 * Responsible for handling threaded parallel malware scan.
 * 
 * Keeps track of a set of scanners and records which have replied. 
 * 
 * @author Jran Vagnby Lillesand
 * @see Scanner#run()
 * 
 */
public class ThreadHandler implements ScanThreadCallback, AVUpdateThreadCallback{

	private ArrayList<ScanResult> scanResults;
	private ArrayList<AVUpdateResult> avUpdateResults; 
	
	private ArrayList<Scanner> unfinishedScanners;
	private ArrayList<Scanner> scanners;
	private boolean running = false; 
	
	/**
	 * Instantiates a ThreadHandler.
	 * 
	 * @param scanners the list of scanners to be used for scan operations.
	 */
	public ThreadHandler(ArrayList<Scanner> scanners){
		this.scanners = scanners;
	}
	
	public void reportResult(ScanResult sr) {
		if(running){
			synchronized(unfinishedScanners){
				scanResults.add(sr);
				unfinishedScanners.remove(sr.getScanner());
			}
		}
	}
	
	public void reportResult(AVUpdateResult rsr){
		if(running){
			synchronized(unfinishedScanners){
				avUpdateResults.add(rsr);
				unfinishedScanners.remove(rsr.getScanner());
			}
		}
	}

	
	/**
	 * Update virus definitions for the scanners associated with this thread handler.
	 * 
	 * Starts the update threads and waits for callbacks until some timeout given in the settings file.
	 * 
	 * @return list over scan results.
	 * @see Scanner#updateAV()
	 * @see Scanner#startThreadedAVUpdate(AVUpdateThreadCallback)
	 */
	public ArrayList<AVUpdateResult> updateScanners(){
		// Only one thread may use the scanners at a time.
		synchronized(scanners){
			//populate list of unfinished scanners
			unfinishedScanners = (ArrayList<Scanner>) scanners.clone();
			
			avUpdateResults = new ArrayList<AVUpdateResult>();
			
			running = true;
			
			//initiate operation in all scanners
			for(Scanner scanner: scanners){
				scanner.startThreadedAVUpdate(this);
			}
			
			// Get constants related to updating
			int timeout = PropertiesReader.getInstance().getInt("update.timeout", 600);
			double pollingInterval = PropertiesReader.getInstance().getDouble("polling.interval.major", 1.000);
			
			// Poll to see if all update operations are done
			for(int i = 0; i < timeout / pollingInterval; i++){
				try{
					Thread.sleep((long) (pollingInterval*1000));
				}catch(InterruptedException e){}
				
				synchronized(unfinishedScanners){
					if(unfinishedScanners.isEmpty()){
						break;
					}
				}
			}
			
			// don't accept any new results (to prevent concurrency issues)
			synchronized(unfinishedScanners){
				running = false;
			}
			
			for(Scanner scanner: unfinishedScanners){
				AVUpdateResult result = new AVUpdateResult(scanner);
				result.setResultState(AVUpdateResult.OPERATION_TIMEOUT);
				
				//TODO: Allow thread to finish, although it might have crashed. Will have to be reviewed by user.
			}
			
			return avUpdateResults;
		}
	}

	/**
	 * Starts scanner threads which executes a malware and attempts to read the real-time anti-virus log file.
	 * 
	 *  Will wait for scan threads to call back either until all threads have completed or timeout as described in the properties file occurs.
	 *  Any threads which fails to report back before the timeout will have their virtual machines forcefully restored to snapshots. This currently causes
	 *  the thread to hang (as it will usually be in a blocking call against the vix api, which is then interrupted by the restore to snapshot instruction).
	 *  Scan results are generated for timed out scan engines. 
	 * 
	 * @param malwareSample referece to the file to be executed.
	 * @return list of real-time scan results.
	 * @see Scanner#executeMalwareSample(File, boolean, boolean)
	 * @see Scanner#startThreadedMalwareExecution(File, boolean, ScanThreadCallback)
	 */
	public ArrayList<ScanResult> executeMalware(File malwareSample){
		// Only one thread may use the scanners at a time.
		synchronized(scanners){
			//Populate list of unfinished scanners
			unfinishedScanners = (ArrayList<Scanner>) scanners.clone();
			
			scanResults = new ArrayList<ScanResult>();
			
			running = true;
			
			// Start execution in all registered scanners (scanners not supporting real-time scanning will simply return at once, so no need to differentiate).
			for(Scanner scanner:scanners){
				scanner.startThreadedMalwareExecution(malwareSample, PropertiesReader.getInstance().getBool("snapshot.before.scan", false), this);
			}
			
			// Get constants related to timing of malware execution
			int timeout = PropertiesReader.getInstance().getInt("full.execution.timeout", 240);
			double pollingInterval = PropertiesReader.getInstance().getDouble("polling.interval.major", 1.000);
			
			for(int i = 0; i < timeout/pollingInterval; i++){
				try{
					Thread.sleep((long) (pollingInterval * 1000));
				}catch(InterruptedException e){}
				
				if(unfinishedScanners.isEmpty()){
					break;
				}
			}
			
			// Don't accept new results.
			synchronized(unfinishedScanners){
				running = false;
			}
			
			for(Scanner scanner: unfinishedScanners){
				ScanResult sr = new ScanResult(scanner);
				
				//revert scanner to snapshot
				
				try{
					scanner.revertToSnapshot(false);
				}catch(Exception e){
					// Okay to catch broad here, the result is the same no matter the exception.
					Frontend.getInstance().displayError("Virtual machine running "+scanner.getName()+" failed to revert to snapshot. Please ensure that the VM is in a ready state before performing a new scan. NOTE THAT MALWARE MAY STILL BE RUNNING ON THE VIRTUAL MACHINE. Error: "+e.getMessage());
				}
				
				sr.setScanTime(timeout);
				sr.setResult("Malware execution timed out due to global timeout.");
				scanResults.add(sr);
			}
			
			return scanResults;
		}
		
	}

	/**
	 * Starts scan threads and waits for them to report back.
	 * 
	 *  Will wait for scan threads to call back either until all threads have completed or timeout as described in the properties file occurs.
	 *  Any threads which fails to report back before the timeout will have their virtual machines forcefully restored to snapshots. This currently causes
	 *  the thread to hang (as it will usually be in a blocking call against the vix api, which is then interrupted by the restore to snapshot instruction).
	 *  Scan results are generated for timed out scan engines. 
	 * 
	 * @param malwareSample malware sample to be scanned.
	 * @return list of scan results.
	 * @see Scanner#scan(File, boolean, boolean)
	 * @see Scanner#startThreadedScan(File, boolean, ScanThreadCallback)
	 */
	public ArrayList<ScanResult> performScan(File malwareSample){
		// Only one thread may use the scanners at a time.
		synchronized(scanners){
			
			//populate list of unfinished scanners
			unfinishedScanners = (ArrayList<Scanner>) scanners.clone();
			
			scanResults = new ArrayList<ScanResult>();
			
			running = true;
			
			//start all scanners for the given file
			for(Scanner scanner: scanners){
				scanner.startThreadedScan(malwareSample, PropertiesReader.getInstance().getBool("snapshot.before.scan", false), this);
			}
			
			// Get constants related to scanning
			int timeout = PropertiesReader.getInstance().getInt("full.scan.timeout", 240);
			double pollingInterval = PropertiesReader.getInstance().getDouble("polling.interval.major", 1.000);
			
			// Poll to see if all scans are done every polling interval:
			for(int i = 0; i < timeout / pollingInterval; i++){
				try{
					Thread.sleep((long) (pollingInterval*1000));
				}catch(InterruptedException e){}
				
				synchronized(unfinishedScanners){
					if(unfinishedScanners.isEmpty()){
						break;
					}
				}
			}
			
			// don't accept any new results (to prevent concurrency issues)
			synchronized(unfinishedScanners){
				running = false;
			}
			
			
			for(Scanner scanner: unfinishedScanners){
				ScanResult sr = new ScanResult(scanner);
				
				try{
					scanner.revertToSnapshot(false);
				}catch(Exception e){
					// Will not happen as call is not made blocking... 
					Frontend.getInstance().displayError("Virtual machine running "+scanner.getName()+" failed to revert to snapshot. Please ensure that the VM is in a ready state before performing a new scan. Error: "+e.getMessage());
				}
				
				
				sr.setScanTime(timeout);
				sr.setResult("Did not respond in time (full operation timeout).");
				
				
				scanResults.add(sr);			
			}
			
			
			return scanResults;
		}
	}
}
